# vim:sw=4:ts=4:et:
"""Python Ring RingGeneric wrapper."""

# pylint: disable=useless-object-inheritance
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar

import pytz

from ring_doorbell.const import URL_DOORBELL_HISTORY, RingCapability
from ring_doorbell.util import (
    parse_datetime,
)

_LOGGER = logging.getLogger(__name__)


class RingGeneric:
    """Generic Implementation for Ring Chime/Doorbell."""

    if TYPE_CHECKING:
        from ring_doorbell.ring import Ring

    def __init__(self, ring: Ring, device_api_id: int) -> None:
        """Initialize Ring Generic."""
        self._ring = ring
        # This is the account ID of the device.
        # Not the same as device ID.
        self.device_api_id = device_api_id
        self.capability = False
        self.alert = None
        self._health_attrs: dict[str, Any] = {}
        self._last_history: list[dict[str, Any]] = []

        # alerts notifications
        self.alert_expires_at = None

    def __repr__(self) -> str:
        """Return __repr__."""
        return f"<{self.__class__.__name__}: {self.name}>"

    def __str__(self) -> str:
        """Return string representation of device."""
        return f"{self.name} ({self.kind})"

    async def async_update(self) -> None:
        """Update this device info."""
        await self.async_update_health_data()

    async def async_update_health_data(self) -> None:
        """Update the health data."""
        raise NotImplementedError

    @property
    def _attrs(self) -> dict[str, Any]:
        """Return attributes."""
        return self._ring.devices_data[self.family][self.device_api_id]

    @property
    def id(self) -> int:
        """Return ID."""
        return self.device_api_id

    @property
    def name(self) -> str:
        """Return name."""
        return self._attrs["description"]

    @property
    def device_id(self) -> str:
        """Return device ID.

        This is the device_id returned by the api, usually the MAC.
        Not to be confused with the id for the device
        """
        return self._attrs["device_id"]

    @property
    def location_id(self) -> str | None:
        """Return location id."""
        return self._attrs.get("location_id", None)

    @property
    def family(self) -> str:
        """Return Ring device family type."""
        raise NotImplementedError

    @property
    def model(self) -> str:
        """Return Ring device model name."""
        raise NotImplementedError

    @property
    def battery_life(self) -> int | None:
        """Return battery life."""
        raise NotImplementedError

    def has_capability(self, capability: RingCapability | str) -> bool:  # noqa: ARG002
        """Return if device has specific capability."""
        return self.capability

    @property
    def address(self) -> str | None:
        """Return address."""
        return self._attrs.get("address")

    @property
    def firmware(self) -> str | None:
        """Return firmware."""
        return self._attrs.get("firmware_version")

    @property
    def latitude(self) -> float | None:
        """Return latitude attr."""
        return self._attrs.get("latitude")

    @property
    def longitude(self) -> float | None:
        """Return longitude attr."""
        return self._attrs.get("longitude")

    @property
    def kind(self) -> str:
        """Return kind attr."""
        return self._attrs["kind"]

    @property
    def timezone(self) -> str | None:
        """Return timezone."""
        return self._attrs.get("time_zone")

    @property
    def wifi_name(self) -> str | None:
        """Return wifi ESSID name.

        Requires health data to be updated.
        """
        return self._health_attrs.get("wifi_name")

    @property
    def wifi_signal_strength(self) -> int | None:
        """Return wifi RSSI.

        Requires health data to be updated.
        """
        return self._health_attrs.get("latest_signal_strength")

    @property
    def wifi_signal_category(self) -> str | None:
        """Return wifi signal category.

        Requires health data to be updated.
        """
        return self._health_attrs.get("latest_signal_category")

    @property
    def last_history(self) -> list[dict[str, Any]]:
        """Return the result of the last history query."""
        return self._last_history

    async def async_history(  # noqa: C901, PLR0913, PLR0912
        self,
        *,
        limit: int = 30,
        timezone: str | None = None,
        kind: str | None = None,
        enforce_limit: bool = False,
        older_than: int | None = None,
        retry: int = 8,
        convert_timezone: bool = True,
    ) -> list[dict[str, Any]]:
        """
        Return history with datetime objects.

        :param limit: specify number of objects to be returned
        :param timezone: determine which timezone to convert data objects
        :param kind: filter by kind (ding, motion, on_demand)
        :param enforce_limit: when True, this will enforce the limit and kind
        :param older_than: return older objects than the passed event_id
        :param retry: determine the max number of attempts to archive the limit
        """
        if not self.has_capability("history"):
            return []

        queries = 0
        original_limit = limit

        # set cap for max queries
        # pylint:disable=consider-using-min-builtin
        retry = min(retry, 10)

        while True:
            params = {"limit": limit}
            if older_than:
                params["older_than"] = older_than

            url = URL_DOORBELL_HISTORY.format(self.device_api_id)
            resp = await self._ring.async_query(url, extra_params=params)
            response = resp.json()

            # cherrypick only the selected kind events
            if kind:
                response = list(filter(lambda array: array["kind"] == kind, response))

            if convert_timezone:
                # convert for specific timezone
                if timezone:
                    mytz = pytz.timezone(timezone)

                for entry in response:
                    utc_dt = parse_datetime(entry["created_at"])
                    if timezone:
                        tz_dt = utc_dt.astimezone(mytz)
                        entry["created_at"] = tz_dt
                    else:
                        entry["created_at"] = utc_dt

            if enforce_limit:
                # return because already matched the number
                # of events by kind
                if len(response) >= original_limit:
                    return response[:original_limit]

                # ensure the loop will exit after max queries
                queries += 1
                if queries == retry:
                    _LOGGER.debug(
                        "Could not find total of %s of kind %s", original_limit, kind
                    )
                    break

                # ensure the kind objects returned to match limit
                limit = limit * 2

            else:
                break

        self._last_history = response
        return self._last_history

    DEPRECATED_API_QUERIES: ClassVar = {
        "history",
        "update",
        "update_health_data",
    }
    DEPRECATED_API_PROPERTY_GETTERS: ClassVar[set[str]] = set()
    DEPRECATED_API_PROPERTY_SETTERS: ClassVar[set[str]] = set()

    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Any:
            """Get a deprecated attribute or raise an error."""
            if name in self.DEPRECATED_API_QUERIES:
                return self._ring.auth._dep_handler.get_api_query(self, name)  # noqa: SLF001
            if name in self.DEPRECATED_API_PROPERTY_GETTERS:
                return self._ring.auth._dep_handler.get_api_property(self, name)  # noqa: SLF001
            msg = f"{self.__class__.__name__} has no attribute {name!r}"
            raise AttributeError(msg)

        def __setattr__(self, name: str, value: Any) -> None:
            """Set a deprecated attribute or raise an error."""
            if name in self.DEPRECATED_API_PROPERTY_SETTERS:
                self._ring.auth._dep_handler.set_api_property(self, name, value)  # noqa: SLF001
            else:
                super().__setattr__(name, value)



"""ring-doorbell exceptions."""


class RingError(Exception):
    """Base exception for device errors."""


class Requires2FAError(RingError):
    """Exception that 2FA is required."""


class AuthenticationError(RingError):
    """Exception for ring authentication errors."""


class RingTimeout(RingError):  # noqa: N818
    """Exception for ring authentication errors."""



"""Python Package for interacting with Ring devices."""

from importlib.metadata import version

__version__ = version("ring_doorbell")

from ring_doorbell.auth import Auth
from ring_doorbell.chime import RingChime
from ring_doorbell.const import RingCapability, RingEventKind
from ring_doorbell.doorbot import RingDoorBell
from ring_doorbell.event import RingEvent
from ring_doorbell.exceptions import (
    AuthenticationError,
    Requires2FAError,
    RingError,
    RingTimeout,
)
from ring_doorbell.generic import RingGeneric
from ring_doorbell.group import RingLightGroup
from ring_doorbell.listen import RingEventListener, RingEventListenerConfig
from ring_doorbell.other import RingOther
from ring_doorbell.ring import Ring, RingDevices
from ring_doorbell.stickup_cam import RingStickUpCam

__all__ = [
    "Ring",
    "Auth",
    "RingDevices",
    "RingChime",
    "RingCapability",
    "RingEventKind",
    "RingStickUpCam",
    "RingLightGroup",
    "RingDoorBell",
    "RingOther",
    "RingEvent",
    "RingEventListener",
    "RingEventListenerConfig",
    "RingError",
    "AuthenticationError",
    "Requires2FAError",
    "RingTimeout",
    "RingGeneric",
    "RingEvent",
]



"""Test configuration for the Ring platform."""

from __future__ import annotations

import datetime
import json
import re
from pathlib import Path
from time import time
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest
from aioresponses import CallbackResult, aioresponses
from ring_doorbell import Auth, Ring
from ring_doorbell.const import USER_AGENT

if TYPE_CHECKING:
    from collections.abc import Generator


# The kwargs below are useful for request assertions
def json_request_kwargs():
    return {
        "headers": {
            "User-Agent": "android:com.ringapp",
            "hardware_id": "21ac3af1-0eac-5fbd-8b0f-0b784889bfbd",
            "Content-Type": "application/json",
            "Authorization": "Bearer dummyBearerToken",
        },
        "timeout": 10,
        "data": None,
        "params": {},
        "json": {},
    }


def nojson_request_kwargs():
    return {
        "headers": {
            "User-Agent": "android:com.ringapp",
            "hardware_id": "21ac3af1-0eac-5fbd-8b0f-0b784889bfbd",
            "Authorization": "Bearer dummyBearerToken",
        },
        "timeout": 10,
        "data": None,
        "params": {},
    }


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "nolistenmock: mark test to not want the autouse listenmock"
    )


@pytest.fixture
async def auth():
    """Return auth object."""
    auth = Auth(USER_AGENT)
    await auth.async_fetch_token("foo", "bar")
    yield auth

    await auth.async_close()


@pytest.fixture
async def ring(auth):
    """Return updated ring object."""
    ring = Ring(auth)
    await ring.async_update_data()
    return ring


def _set_dings_to_now(active_dings) -> None:
    for ding in active_dings:
        ding["now"] = time()

    return active_dings


def load_fixture(filename):
    """Load a fixture."""
    path = Path(Path(__file__).parent / "fixtures" / filename)
    with path.open() as fdp:
        return fdp.read()


def load_fixture_as_dict(filename):
    """Load a fixture."""
    return json.loads(load_fixture(filename))


def load_alert_v1(
    alert_type: str, device_id, *, ding_id_inc: int = 0, created_at: str | None = None
) -> dict:
    msg = json.loads(load_fixture(Path().joinpath("listen", "fcmdata_v1.json")))
    gcmdata = json.loads(
        load_fixture(Path().joinpath("listen", f"{alert_type}_gcmdata.json"))
    )
    if created_at is None:
        created_at = datetime.datetime.utcnow().isoformat(timespec="milliseconds") + "Z"  # noqa: DTZ003
    if "ding" in gcmdata:
        gcmdata["ding"]["doorbot_id"] = device_id
        gcmdata["ding"]["created_at"] = created_at
        gcmdata["ding"]["id"] = gcmdata["ding"]["id"] + ding_id_inc
    else:
        gcmdata["alarm_meta"]["device_zid"] = device_id
    msg["data"]["gcmData"] = json.dumps(gcmdata)
    return msg


def load_alert_v2(
    alert_type: str, device_id, *, ding_id_inc: int = 0, created_at: str | None = None
) -> dict:
    msg = json.loads(load_fixture(Path().joinpath("listen", "fcmdata_v2.json")))
    data = json.loads(
        load_fixture(Path().joinpath("listen", f"{alert_type}_data.json"))
    )
    android_config = json.loads(
        load_fixture(Path().joinpath("listen", f"{alert_type}_android_config.json"))
    )
    analytics = json.loads(
        load_fixture(Path().joinpath("listen", f"{alert_type}_analytics.json"))
    )
    if created_at is None:
        created_at = datetime.datetime.utcnow().isoformat(timespec="milliseconds") + "Z"  # noqa: DTZ003
    data["device"]["id"] = device_id
    data["event"]["ding"]["created_at"] = created_at
    data["event"]["ding"]["id"] = str(int(data["event"]["ding"]["id"]) + ding_id_inc)
    msg["data"]["data"] = json.dumps(data)
    msg["data"]["android_config"] = json.dumps(android_config)
    msg["data"]["analytics"] = json.dumps(analytics)
    return msg


@pytest.fixture(autouse=True)
def _listen_mock(mocker, request) -> None:
    if "nolistenmock" in request.keywords:
        return

    mocker.patch(
        "firebase_messaging.FcmPushClient.checkin_or_register", return_value="foobar"
    )
    mocker.patch("firebase_messaging.FcmPushClient.start")
    mocker.patch("firebase_messaging.FcmPushClient.stop")
    mocker.patch("firebase_messaging.FcmPushClient.is_started", return_value=True)


def callback(url, **kwargs):  # noqa: ANN003
    return CallbackResult(status=418)


# tests to pull in request_mock and append uris
@pytest.fixture
def devices_fixture():
    class Devices:
        def __init__(self) -> None:
            """Initialise the class."""
            self.updated = False

        def devices(self) -> dict:
            """Get the devices."""
            if not self.updated:
                return load_fixture_as_dict("ring_devices.json")
            return load_fixture_as_dict("ring_devices_updated.json")

        def callback(self, url, **kwargs) -> CallbackResult:  # noqa: ARG002, ANN003
            """Return the callback result."""
            return CallbackResult(payload=self.devices())

    return Devices()


@pytest.fixture
def putpatch_status_fixture():
    class StatusOverrides:
        def __init__(self) -> None:
            """Initialise the class."""
            self.overrides = {}

        def callback(self, url, **kwargs) -> CallbackResult:  # noqa: ANN003, ARG002
            """Return the callback."""
            plain_url = str(url)
            if plain_url in self.overrides:
                return CallbackResult(body=b"", status=self.overrides[plain_url])

            return CallbackResult(body=b"", status=204)

    return StatusOverrides()


@pytest.fixture(autouse=True, name="hardware_id_mock")
def _hardware_id_mock_fixture() -> Generator:
    """Fixture to patch getnode ensures all tests generate same hardware_id."""
    with patch("uuid.getnode", return_value=12345678901):
        yield


# setting the fixture name to requests_mock allows other
# tests to pull in request_mock and append uris
@pytest.fixture(autouse=True, name="aioresponses_mock")
def aioresponses_mock_fixture(request, devices_fixture, putpatch_status_fixture):
    with aioresponses() as mock:
        mock.post(
            "https://oauth.ring.com/oauth/token",
            payload=load_fixture_as_dict("ring_oauth.json"),
            repeat=True,
        )
        mock.post(
            "https://api.ring.com/clients_api/session",
            payload=load_fixture_as_dict("ring_session.json"),
            repeat=True,
        )
        mock.get(
            "https://api.ring.com/clients_api/ring_devices",
            callback=devices_fixture.callback,
            repeat=True,
        )
        mock.get(
            re.compile(r"https:\/\/api\.ring\.com\/clients_api\/chimes\/\d+\/health"),
            payload=load_fixture_as_dict("ring_chime_health_attrs.json"),
            repeat=True,
        )
        mock.get(
            re.compile(r"https:\/\/api\.ring\.com\/clients_api\/doorbots\/\d+\/health"),
            payload=load_fixture_as_dict("ring_doorboot_health_attrs.json"),
            repeat=True,
        )
        mock.get(
            re.compile(
                r"https:\/\/api\.ring\.com\/clients_api\/doorbots\/185036587\/history.*$"
            ),
            payload=load_fixture_as_dict("ring_intercom_history.json"),
            repeat=True,
        )
        mock.get(
            re.compile(
                r"https:\/\/api\.ring\.com\/clients_api\/doorbots\/\d+\/history.*$"
            ),
            payload=load_fixture_as_dict("ring_doorbot_history.json"),
            repeat=True,
        )
        mock.get(
            "https://api.ring.com/clients_api/dings/active",
            payload=_set_dings_to_now(load_fixture_as_dict("ring_ding_active.json")),
            repeat=True,
        )
        mock.put(
            "https://api.ring.com/clients_api/doorbots/987652/floodlight_light_off",
            payload="ok",
            repeat=True,
        )
        mock.put(
            "https://api.ring.com/clients_api/doorbots/987652/floodlight_light_on",
            payload="ok",
            repeat=True,
        )
        mock.put(
            "https://api.ring.com/clients_api/doorbots/987652/siren_on",
            payload="ok",
            repeat=True,
        )
        mock.put(
            "https://api.ring.com/clients_api/doorbots/987652/siren_off",
            payload="ok",
            repeat=True,
        )
        mock.get(
            "https://api.ring.com/groups/v1/locations/mock-location-id/groups",
            payload=load_fixture_as_dict("ring_groups.json"),
            repeat=True,
        )
        mock.get(
            "https://api.ring.com/groups/v1/locations/"
            "mock-location-id/groups/mock-group-id/devices",
            payload=load_fixture_as_dict("ring_group_devices.json"),
            repeat=True,
        )
        mock.post(
            "https://api.ring.com/groups/v1/locations/"
            "mock-location-id/groups/mock-group-id/devices",
            payload="ok",
            repeat=True,
        )
        mock.patch(
            re.compile(
                r"https:\/\/api\.ring\.com\/devices\/v1\/devices\/\d+\/settings"
            ),
            payload="ok",
            repeat=True,
        )
        mock.get(
            re.compile(r"https:\/\/api\.ring\.com\/clients_api\/dings\/\d+\/recording"),
            status=200,
            body=b"123456",
            repeat=True,
        )
        mock.get(
            "https://api.ring.com/clients_api/dings/9876543212/recording",
            status=200,
            body=b"123456",
            repeat=True,
        )
        mock.patch(
            "https://api.ring.com/clients_api/device",
            callback=putpatch_status_fixture.callback,
            repeat=True,
        )
        mock.put(
            re.compile(r"https:\/\/api\.ring\.com\/clients_api\/doorbots\/.*$"),
            status=204,
            body=b"",
            repeat=True,
        )
        mock.get(
            "https://api.ring.com/devices/v1/devices/185036587/settings",
            payload=load_fixture_as_dict("ring_intercom_settings.json"),
            repeat=True,
        )
        mock.get(
            "https://api.ring.com/clients_api/locations/mock-location-id/users",
            payload=load_fixture_as_dict("ring_intercom_users.json"),
            repeat=True,
        )
        mock.post(
            "https://api.ring.com/clients_api/locations/mock-location-id/invitations",
            payload="ok",
            repeat=True,
        )
        mock.delete(
            (
                "https://api.ring.com/clients_api/locations/"
                "mock-location-id/invitations/123456789"
            ),
            payload="ok",
            repeat=True,
        )
        requestid = "44529542-3ed7-41da-807e-c170a01bac1d"
        mock.put(
            "https://api.ring.com/commands/v1/devices/185036587/device_rpc",
            body=(
                '{"result": {"code": 0}, "id": "' + requestid + '", "jsonrpc": "2.0"}'
            ).encode(),
            repeat=True,
        )
        yield mock



"""The tests for the Ring platform."""

from .conftest import json_request_kwargs, nojson_request_kwargs


async def test_other_attributes(ring):
    """Test the Ring Other class and methods."""
    dev = ring.devices()["other"][0]

    assert dev.id != 99999
    assert dev.device_id == "124ba1b3fe1a"
    assert dev.kind == "intercom_handset_audio"
    assert dev.model == "Intercom"
    assert dev.location_id == "mock-location-id"
    assert dev.has_capability("battery") is False
    assert dev.has_capability("open") is True
    assert dev.has_capability("history") is True
    assert dev.timezone == "Europe/Rome"
    assert dev.battery_life == 52
    assert dev.doorbell_volume == 8
    assert dev.mic_volume == 11
    assert await dev.async_get_clip_length_max() == 60
    assert dev.connection_status == "online"
    assert len(await dev.async_get_allowed_users()) == 2
    assert dev.subscribed is True
    assert dev.has_subscription is True
    assert dev.unlock_duration is None
    assert dev.keep_alive_auto == 45.0

    assert isinstance(await dev.async_history(limit=1, kind="on_demand"), list)
    assert len(await dev.async_history(kind="ding")) == 1
    assert len(await dev.async_history(limit=1, kind="on_demand")) == 2
    assert (
        len(
            await dev.async_history(
                limit=1, kind="on_demand", enforce_limit=True, retry=50
            )
        )
        == 1
    )

    await dev.async_update_health_data()
    assert dev.wifi_name == "ring_mock_wifi"
    assert dev.wifi_signal_category == "good"
    assert dev.wifi_signal_strength != 100


async def test_other_controls(ring, aioresponses_mock):
    dev = ring.devices()["other"][0]

    kwargs = json_request_kwargs()
    kwargs["json"] = None

    await dev.async_set_doorbell_volume(6)
    kwargs["params"] = {"doorbot[settings][doorbell_volume]": "6"}
    aioresponses_mock.assert_called_with(
        "https://api.ring.com/clients_api/doorbots/185036587", method="PUT", **kwargs
    )

    kwargs = json_request_kwargs()

    await dev.async_set_mic_volume(10)
    kwargs["json"] = {"volume_settings": {"mic_volume": 10}}
    aioresponses_mock.assert_called_with(
        "https://api.ring.com/devices/v1/devices/185036587/settings",
        method="PATCH",
        **kwargs,
    )

    await dev.async_set_voice_volume(9)
    kwargs["json"] = {"volume_settings": {"voice_volume": 9}}
    aioresponses_mock.assert_called_with(
        "https://api.ring.com/devices/v1/devices/185036587/settings",
        method="PATCH",
        **kwargs,
    )

    await dev.async_set_clip_length_max(30)
    kwargs["json"] = {"video_settings": {"clip_length_max": 30}}
    aioresponses_mock.assert_called_with(
        "https://api.ring.com/devices/v1/devices/185036587/settings",
        method="PATCH",
        **kwargs,
    )

    await dev.async_set_keep_alive_auto(32.2)
    kwargs["json"] = {"keep_alive_settings": {"keep_alive_auto": 32.2}}
    aioresponses_mock.assert_called_with(
        "https://api.ring.com/devices/v1/devices/185036587/settings",
        method="PATCH",
        **kwargs,
    )


async def test_other_invitations(ring, aioresponses_mock):
    dev = ring.devices()["other"][0]
    kwargs = json_request_kwargs()
    kwargs["json"] = {
        "invitation": {
            "doorbot_ids": [185036587],
            "invited_email": "test@example.com",
            "group_ids": [],
        }
    }

    await dev.async_invite_access("test@example.com")
    aioresponses_mock.assert_called_with(
        "https://api.ring.com/clients_api/locations/mock-location-id/invitations",
        method="POST",
        **kwargs,
    )

    await dev.async_remove_access(123456789)

    kwargs = nojson_request_kwargs()
    aioresponses_mock.assert_called_with(
        "https://api.ring.com/clients_api/locations/mock-location-id/invitations/123456789",
        method="DELETE",
        **kwargs,
    )


async def test_other_open_door(ring, aioresponses_mock, mocker):
    dev = ring.devices()["other"][0]

    mocker.patch("uuid.uuid4", return_value="987654321")

    kwargs = json_request_kwargs()
    kwargs["json"] = {
        "command_name": "device_rpc",
        "request": {
            "id": "987654321",
            "jsonrpc": "2.0",
            "method": "unlock_door",
            "params": {"door_id": 0, "user_id": 15},
        },
    }

    await dev.async_open_door(15)
    aioresponses_mock.assert_called_with(
        "https://api.ring.com/commands/v1/devices/185036587/device_rpc",
        method="PUT",
        **kwargs,
    )



"""The tests for the Ring platform."""

import asyncio
from datetime import datetime, timezone

import pytest
from freezegun.api import FrozenDateTimeFactory
from ring_doorbell import Auth, Ring, RingError
from ring_doorbell.const import MSG_EXISTING_TYPE, USER_AGENT
from ring_doorbell.util import parse_datetime

from .conftest import json_request_kwargs, load_fixture_as_dict


def test_basic_attributes(ring):
    """Test the Ring class and methods."""
    data = ring.devices()
    assert len(data["chimes"]) == 1
    assert len(data["doorbots"]) == 1
    assert len(data["authorized_doorbots"]) == 1
    assert len(data["stickup_cams"]) == 1
    assert len(data["other"]) == 1


async def test_chime_attributes(ring):
    """Test the Ring Chime class and methods."""
    dev = ring.devices()["chimes"][0]

    assert dev.address == "123 Main St"
    assert dev.id != 99999
    assert dev.device_id == "abcdef123"
    assert dev.kind == "chime"
    assert dev.model == "Chime"
    assert dev.has_capability("battery") is False
    assert dev.has_capability("volume") is True
    assert dev.has_capability("history") is False
    assert dev.latitude is not None
    assert dev.timezone == "America/New_York"
    assert dev.volume == 2

    assert len(await dev.async_history()) == 0

    await dev.async_update_health_data()
    assert dev.wifi_name == "ring_mock_wifi"
    assert dev.wifi_signal_category == "good"
    assert dev.wifi_signal_strength != 100


async def test_doorbell_attributes(ring):
    data = ring.devices()
    dev = data["doorbots"][0]
    assert dev.name == "Front Door"
    assert dev.id == 987652
    assert dev.address == "123 Main St"
    assert dev.kind == "lpd_v1"
    assert dev.model == "Doorbell Pro"
    assert dev.has_capability("battery") is False
    assert dev.has_capability("volume") is True
    assert dev.has_capability("history") is True
    assert dev.longitude == -70.12345
    assert dev.timezone == "America/New_York"
    assert dev.volume == 1
    assert dev.has_subscription is True
    assert dev.connection_status == "online"

    assert isinstance(await dev.async_history(limit=1, kind="motion"), list)
    assert len(await dev.async_history(kind="ding")) == 1
    assert len(await dev.async_history(limit=1, kind="motion")) == 2
    assert (
        len(
            await dev.async_history(
                limit=1, kind="motion", enforce_limit=True, retry=50
            )
        )
        == 1
    )

    assert dev.existing_doorbell_type == "Mechanical"

    await dev.async_update_health_data()

    assert dev.wifi_name == "ring_mock_wifi"
    assert dev.wifi_signal_category == "good"
    assert dev.wifi_signal_strength == -58


def test_shared_doorbell_attributes(ring):
    data = ring.devices()
    dev = data["authorized_doorbots"][0]

    assert dev.id == 987653
    assert dev.battery_life == 51
    assert dev.address == "123 Second St"
    assert dev.kind == "lpd_v1"
    assert dev.model == "Doorbell Pro"
    assert dev.has_capability("battery") is False
    assert dev.has_capability("volume") is True
    assert dev.has_capability("history") is True
    assert dev.longitude == -70.12345
    assert dev.timezone == "America/New_York"
    assert dev.volume == 5
    assert dev.existing_doorbell_type == "Digital"


def test_stickup_cam_attributes(ring):
    dev = ring.devices()["stickup_cams"][0]
    assert dev.kind == "hp_cam_v1"
    assert dev.model == "Floodlight Cam"
    assert dev.has_capability("battery") is False
    assert dev.has_capability("light") is True
    assert dev.has_capability("history") is True
    assert dev.lights == "off"
    assert dev.siren == 0


async def test_stickup_cam_controls(ring, aioresponses_mock):
    dev = ring.devices()["stickup_cams"][0]

    kwargs = json_request_kwargs()
    kwargs["json"] = None

    await dev.async_set_lights("off")
    aioresponses_mock.assert_called_with(
        url="https://api.ring.com/clients_api/doorbots/987652/floodlight_light_off",
        method="PUT",
        **kwargs,
    )
    await dev.async_set_lights("on")
    aioresponses_mock.assert_called_with(
        url="https://api.ring.com/clients_api/doorbots/987652/floodlight_light_on",
        method="PUT",
        **kwargs,
    )
    await dev.async_set_siren(0)
    aioresponses_mock.assert_called_with(
        url="https://api.ring.com/clients_api/doorbots/987652/siren_off",
        method="PUT",
        **kwargs,
    )
    await dev.async_set_siren(30)
    kwargs["params"] = {"duration": 30}
    aioresponses_mock.assert_called_with(
        url="https://api.ring.com/clients_api/doorbots/987652/siren_on",
        method="PUT",
        **kwargs,
    )


async def test_light_groups(ring):
    group = ring.groups()["mock-group-id"]

    assert group.name == "Landscape"
    assert group.family == "group"
    assert group.group_id == "mock-group-id"
    assert group.location_id == "mock-location-id"
    assert group.model == "Light Group"
    assert group.has_capability("light") is True
    with pytest.raises(RingError):
        group.has_capability("something-else")

    with pytest.raises(
        RingError,
        match=(
            "You need to call update on the group before "
            "accessing the lights property."
        ),
    ):
        assert group.lights is False
    await group.async_update()

    # Attempt turning on lights
    await group.async_set_lights(state=True)

    # Attempt turning off lights
    await group.async_set_lights(state=False)

    # Attempt turning on lights for 30 seconds
    await group.async_set_lights(state=True, duration=30)


async def test_motion_detection_enable(ring, aioresponses_mock):
    dev = ring.devices()["doorbots"][0]

    kwargs = json_request_kwargs()
    await dev.async_set_motion_detection(state=True)
    kwargs["json"] = {"motion_settings": {"motion_detection_enabled": True}}
    aioresponses_mock.assert_called_with(
        url="https://api.ring.com/devices/v1/devices/987652/settings",
        method="PATCH",
        **kwargs,
    )

    await dev.async_set_motion_detection(state=False)

    kwargs["json"] = {"motion_settings": {"motion_detection_enabled": False}}
    aioresponses_mock.assert_called_with(
        url="https://api.ring.com/devices/v1/devices/987652/settings",
        method="PATCH",
        **kwargs,
    )


@pytest.mark.parametrize(
    ("datetime_string", "expected", "error_in_log"),
    [
        pytest.param(
            "2012-01-15T06:01:01",
            datetime(2012, 1, 14, 5, 5, 5, 123 * 1_000, tzinfo=timezone.utc),
            True,
            id="No timezone",
        ),
        pytest.param(
            "2012-01-15T06:01:01.12Z",
            datetime(2012, 1, 15, 6, 1, 1, 120 * 1_000, tzinfo=timezone.utc),
            False,
            id="Millis",
        ),
        pytest.param(
            "2012-01-15T06:01:01.123456Z",
            datetime(2012, 1, 15, 6, 1, 1, 123456, tzinfo=timezone.utc),
            False,
            id="Micros",
        ),
        pytest.param(
            "2012-01-15T06:01:01Z",
            datetime(2012, 1, 15, 6, 1, 1, 0, tzinfo=timezone.utc),
            False,
            id="No millis",
        ),
    ],
)
def test_datetime_parse(
    freezer: FrozenDateTimeFactory,
    caplog: pytest.LogCaptureFixture,
    datetime_string,
    expected,
    error_in_log,
):
    """Test the datetime parsing."""
    freezer.move_to("2012-01-14T05:05:05.123Z")
    dt = parse_datetime(datetime_string)
    is_error_in_log = (
        f"Unable to parse datetime string {datetime_string}, defaulting to now time"
        in caplog.text
    )
    assert dt == expected
    assert is_error_in_log is error_in_log


async def test_sync_queries_from_event_loop():
    auth = Auth(USER_AGENT, token=load_fixture_as_dict("ring_oauth.json"))
    ring = Ring(auth)

    assert asyncio.get_running_loop()

    msg = (
        "You cannot call deprecated sync function Ring.update_devices "
        "from within a running event loop."
    )
    with pytest.raises(RingError, match=msg):
        ring.update_devices()


async def test_sync_queries_from_executor():
    auth = Auth(USER_AGENT, token=load_fixture_as_dict("ring_oauth.json"))
    ring = Ring(auth)

    loop = asyncio.get_running_loop()

    assert ring.devices_data == {}
    # This will run the query inan executor thread
    msg = "Ring.update_devices is deprecated, use Ring.async_update_devices"
    with pytest.deprecated_call(match=msg):
        await loop.run_in_executor(None, ring.update_devices)

    assert ring.devices_data


def test_sync_queries_with_no_event_loop():
    auth = Auth(USER_AGENT, token=load_fixture_as_dict("ring_oauth.json"))

    ring = Ring(auth)

    assert not ring.devices_data
    msg = "Ring.update_devices is deprecated, use Ring.async_update_devices"
    with pytest.deprecated_call(match=msg):
        ring.update_devices()

    assert ring.devices_data

    with pytest.deprecated_call():
        auth.close()


async def test_set_existing_doorbell_type(ring, aioresponses_mock):
    data = ring.devices()
    dev = data["doorbots"][0]
    assert dev.existing_doorbell_type == "Mechanical"

    kwargs = json_request_kwargs()
    kwargs["json"] = None

    aioresponses_mock.requests.clear()
    # Attempting to turn off the in-home chime
    await dev.async_set_existing_doorbell_type_enabled(value=False)
    kwargs["params"] = {
        "doorbot[description]": dev.name,
        "doorbot[settings][chime_settings][enable]": 0,
    }
    aioresponses_mock.assert_called_with(
        url="https://api.ring.com/clients_api/doorbots/987652",
        method="PUT",
        **kwargs,
    )

    aioresponses_mock.requests.clear()
    # Attempting to turn on the in-home chime
    await dev.async_set_existing_doorbell_type_enabled(value=True)
    kwargs["params"] = {
        "doorbot[description]": dev.name,
        "doorbot[settings][chime_settings][enable]": 1,
    }
    aioresponses_mock.assert_called_with(
        url="https://api.ring.com/clients_api/doorbots/987652",
        method="PUT",
        **kwargs,
    )

    aioresponses_mock.requests.clear()
    # Attempting to set the doorbell type
    await dev.async_set_existing_doorbell_type(2)
    kwargs["params"] = {
        "doorbot[description]": dev.name,
        "doorbot[settings][chime_settings][type]": 2,
    }
    aioresponses_mock.assert_called_with(
        url="https://api.ring.com/clients_api/doorbots/987652",
        method="PUT",
        **kwargs,
    )

    aioresponses_mock.requests.clear()
    # Attempting to set the duration of the in-home chime
    settings = dev._attrs["settings"]["chime_settings"]
    settings["type"] = 1
    assert dev.existing_doorbell_type == "Digital"
    await dev.async_set_existing_doorbell_type_duration(5)
    kwargs["params"] = {
        "doorbot[description]": dev.name,
        "doorbot[settings][chime_settings][duration]": 5,
    }
    aioresponses_mock.assert_called_with(
        url="https://api.ring.com/clients_api/doorbots/987652",
        method="PUT",
        **kwargs,
    )

    # Attempting to enable when no chime present
    settings = dev._attrs["settings"]["chime_settings"]
    settings["type"] = 2
    assert dev.existing_doorbell_type == "Not Present"

    with pytest.raises(RingError, match="In-Home chime is not present."):
        await dev.async_set_existing_doorbell_type_enabled(value=True)

    # Attempting to set the doorbell type to an invalid value
    with pytest.raises(RingError, match=f"value must be in {MSG_EXISTING_TYPE}"):
        await dev.async_set_existing_doorbell_type(4)

    # Attempting to set the doorbell duration to an invalid value
    with pytest.raises(RingError, match=f"Must be within the {0}-{1}."):
        await dev.async_set_existing_doorbell_type_duration(11)



# vim:sw=4:ts=4:et:
"""Python Ring Chime wrapper."""

from __future__ import annotations

import logging
from typing import Any, ClassVar

from ring_doorbell.const import (
    CHIME_KINDS,
    CHIME_PRO_KINDS,
    CHIME_TEST_SOUND_KINDS,
    CHIME_VOL_MAX,
    CHIME_VOL_MIN,
    CHIMES_ENDPOINT,
    HEALTH_CHIMES_ENDPOINT,
    LINKED_CHIMES_ENDPOINT,
    MSG_VOL_OUTBOUND,
    TESTSOUND_CHIME_ENDPOINT,
    RingCapability,
    RingEventKind,
)
from ring_doorbell.exceptions import RingError
from ring_doorbell.generic import RingGeneric

_LOGGER = logging.getLogger(__name__)


class RingChime(RingGeneric):
    """Implementation for Ring Chime."""

    @property
    def family(self) -> str:
        """Return Ring device family type."""
        return "chimes"

    async def async_update_health_data(self) -> None:
        """Update health attrs."""
        resp = await self._ring.async_query(
            HEALTH_CHIMES_ENDPOINT.format(self.device_api_id)
        )
        self._health_attrs = resp.json().get("device_health", {})

    @property
    def model(self) -> str:
        """Return Ring device model name."""
        if self.kind in CHIME_KINDS:
            return "Chime"
        if self.kind in CHIME_PRO_KINDS:
            return "Chime Pro"
        return "Unknown Chime"

    def has_capability(self, capability: RingCapability | str) -> bool:
        """Return if device has specific capability."""
        capability = (
            capability
            if isinstance(capability, RingCapability)
            else RingCapability.from_name(capability)
        )
        return capability == RingCapability.VOLUME

    @property
    def volume(self) -> int:
        """Return the chime volume."""
        return self._attrs["settings"].get("volume", 0)

    async def async_set_volume(self, value: int) -> None:
        """Set the chime volume."""
        if not ((isinstance(value, int)) and (CHIME_VOL_MIN <= value <= CHIME_VOL_MAX)):
            raise RingError(MSG_VOL_OUTBOUND.format(CHIME_VOL_MIN, CHIME_VOL_MAX))

        params = {
            "chime[description]": self.name,
            "chime[settings][volume]": str(value),
        }
        url = CHIMES_ENDPOINT.format(self.device_api_id)
        await self._ring.async_query(url, extra_params=params, method="PUT")

    async def async_get_linked_tree(self) -> dict[str, Any]:
        """Return doorbell data linked to chime."""
        url = LINKED_CHIMES_ENDPOINT.format(self.device_api_id)
        resp = await self._ring.async_query(url)
        return resp.json()

    async def async_test_sound(
        self, kind: RingEventKind | str = RingEventKind.DING
    ) -> bool:
        """Play chime to test sound."""
        kind_str = kind.value if isinstance(kind, RingEventKind) else kind
        if kind_str not in CHIME_TEST_SOUND_KINDS:
            return False
        url = TESTSOUND_CHIME_ENDPOINT.format(self.device_api_id)
        await self._ring.async_query(
            url, method="POST", extra_params={"kind": kind_str}
        )
        return True

    DEPRECATED_API_QUERIES: ClassVar = {
        *RingGeneric.DEPRECATED_API_QUERIES,
        "update_health_data",
        "test_sound",
    }
    DEPRECATED_API_PROPERTY_GETTERS: ClassVar = {
        *RingGeneric.DEPRECATED_API_PROPERTY_GETTERS,
        "linked_tree",
    }
    DEPRECATED_API_PROPERTY_SETTERS: ClassVar = {
        *RingGeneric.DEPRECATED_API_PROPERTY_SETTERS,
        "volume",
    }



# vim:sw=4:ts=4:et:
"""Constants and enums."""

from __future__ import annotations

from enum import Enum, auto
from typing import Final

from ring_doorbell.exceptions import RingError


class OAuth:
    """OAuth class constants."""

    ENDPOINT = "https://oauth.ring.com/oauth/token"
    CLIENT_ID = "ring_official_android"
    SCOPE: Final[list[str]] = ["client"]


class RingEventKind(Enum):
    """Enum of available ring events."""

    DING = "ding"
    MOTION = "motion"
    INTERCOM_UNLOCK = "intercom_unlock"


class RingCapability(Enum):
    """Enum of available ring events."""

    VIDEO = auto()
    MOTION_DETECTION = auto()
    HISTORY = auto()
    LIGHT = auto()
    SIREN = auto()
    VOLUME = auto()
    BATTERY = auto()
    OPEN = auto()
    KNOCK = auto()
    PRE_ROLL = auto()
    DING = auto()

    @staticmethod
    def from_name(name: str) -> RingCapability:
        """Return ring capability from string value."""
        capability = name.replace("-", "_").upper()
        for ring_capability in RingCapability:
            if ring_capability.name == capability:
                return ring_capability
        msg = f"Unknown ring capability {name}"
        raise RingError(msg)


PACKAGE_NAME = "ring_doorbell"
# timeout for HTTP requests
TIMEOUT = 10

# longer default timeout for recording downloads - typical video file sizes
# are ~12 MB and empirical testing reveals a ~20 second download time over a
# fast connection, suggesting speed is largely governed by capacity of Ring
# backend; to be safe, we factor in a worst case overhead and set it to 2
# minutes (this default can be overridden in method call)
DEFAULT_VIDEO_DOWNLOAD_TIMEOUT = 120


# API endpoints
API_VERSION = "11"
API_URI = "https://api.ring.com"
APP_API_URI = "https://prd-api-us.prd.rings.solutions"
USER_AGENT = "android:com.ringapp"

# random uuid, used to make a hardware id that doesn't change or clash
NAMESPACE_UUID = "379378b0-f747-4b67-a10f-3b13327e8879"

DEFAULT_LISTEN_EVENT_EXPIRES_IN = 180
# for Ring android app.  703521446232 for ring-site
FCM_RING_SENDER_ID = "876313859327"
FCM_API_KEY = "AIzaSyCv-hdFBmmdBBJadNy-TFwB-xN_H5m3Bk8"
FCM_PROJECT_ID = "ring-17770"
FCM_APP_ID = "1:876313859327:android:e10ec6ddb3c81f39"

CLI_TOKEN_FILE = "ring_token.cache"  # noqa: S105
GCM_TOKEN_FILE = "ring_gcm_token.cache"  # noqa: S105
CHIMES_ENDPOINT = "/clients_api/chimes/{0}"
DEVICES_ENDPOINT = "/clients_api/ring_devices"
DINGS_ENDPOINT = "/clients_api/dings/active"
DOORBELLS_ENDPOINT = "/clients_api/doorbots/{0}"
PERSIST_TOKEN_ENDPOINT = "/clients_api/device"  # noqa: S105
SUBSCRIPTION_ENDPOINT = "/clients_api/device"
GROUPS_ENDPOINT = "/groups/v1/locations/{0}/groups"
LOCATIONS_HISTORY_ENDPOINT = "/evm/v2/history/locations/{0}"
LOCATIONS_ENDPOINT = "/clients_api/locations/{0}"

HEALTH_DOORBELL_ENDPOINT = DOORBELLS_ENDPOINT + "/health"
HEALTH_CHIMES_ENDPOINT = CHIMES_ENDPOINT + "/health"
LIGHTS_ENDPOINT = DOORBELLS_ENDPOINT + "/floodlight_light_{1}"
LINKED_CHIMES_ENDPOINT = CHIMES_ENDPOINT + "/linked_doorbots"
LIVE_STREAMING_ENDPOINT = DOORBELLS_ENDPOINT + "/live_view"
NEW_SESSION_ENDPOINT = "/clients_api/session"
RINGTONES_ENDPOINT = "/ringtones"
SIREN_ENDPOINT = DOORBELLS_ENDPOINT + "/siren_{1}"
SNAPSHOT_ENDPOINT = "/clients_api/snapshots/image/{0}"
SNAPSHOT_TIMESTAMP_ENDPOINT = "/clients_api/snapshots/timestamps"
TESTSOUND_CHIME_ENDPOINT = CHIMES_ENDPOINT + "/play_sound"
URL_DOORBELL_HISTORY = DOORBELLS_ENDPOINT + "/history"
URL_RECORDING = "/clients_api/dings/{0}/recording"
URL_RECORDING_SHARE_PLAY = "/clients_api/dings/{0}/share/play"
GROUP_DEVICES_ENDPOINT = GROUPS_ENDPOINT + "/{1}/devices"
SETTINGS_ENDPOINT = "/devices/v1/devices/{0}/settings"
# Alternative API for Intercom history, not used in favor of the DoorBell API
URL_INTERCOM_HISTORY = LOCATIONS_HISTORY_ENDPOINT + "?ringtercom"
INTERCOM_OPEN_ENDPOINT = "/commands/v1/devices/{0}/device_rpc"
INTERCOM_INVITATIONS_ENDPOINT = LOCATIONS_ENDPOINT + "/invitations"
INTERCOM_INVITATIONS_DELETE_ENDPOINT = LOCATIONS_ENDPOINT + "/invitations/{1}"
INTERCOM_ALLOWED_USERS = LOCATIONS_ENDPOINT + "/users"

# New API endpoints for web rtc streaming
RTC_STREAMING_TICKET_ENDPOINT = "/api/v1/clap/ticket/request/signalsocket"
RTC_STREAMING_WEB_SOCKET_ENDPOINT = "wss://api.prod.signalling.ring.devices.a2z.com:443/ws?api_version=4.0&auth_type=ring_solutions&client_id=ring_site-{0}&token={1}"

KIND_DING = "ding"
KIND_MOTION = "motion"
KIND_INTERCOM_UNLOCK = "intercom_unlock"
KIND_ALARM_MODE_NONE = "alarm_mode_none"
KIND_ALARM_MODE_SOME = "alarm_mode_some"
KIND_ALARM_SIREN = "alarm_siren"
KIND_ALARM_SILENCED = "alarm_silenced"

# chime test sound kinds
CHIME_TEST_SOUND_KINDS = (KIND_DING, KIND_MOTION)

# default values
CHIME_VOL_MIN = 0
CHIME_VOL_MAX = 11

DOORBELL_VOL_MIN = 0
DOORBELL_VOL_MAX = 11

MIC_VOL_MIN = 0
MIC_VOL_MAX = 11

VOICE_VOL_MIN = 0
VOICE_VOL_MAX = 11

OTHER_DOORBELL_VOL_MIN = 0
OTHER_DOORBELL_VOL_MAX = 8

DOORBELL_EXISTING_TYPE = {0: "Mechanical", 1: "Digital", 2: "Not Present"}

SIREN_DURATION_MIN = 0
SIREN_DURATION_MAX = 120

DOORBELL_EXISTING_DURATION_MIN = 0
DOORBELL_EXISTING_DURATION_MAX = 10

# device model kinds
CHIME_KINDS = ["chime", "chime_v2"]
CHIME_PRO_KINDS = ["chime_pro", "chime_pro_v2"]

DOORBELL_KINDS = ["doorbot", "doorbell", "doorbell_v3"]
DOORBELL_2_KINDS = ["doorbell_v4", "doorbell_v5"]
DOORBELL_3_KINDS = ["doorbell_scallop_lite"]
DOORBELL_4_KINDS = ["doorbell_oyster"]  # Added
DOORBELL_3_PLUS_KINDS = ["doorbell_scallop"]
DOORBELL_PRO_KINDS = ["lpd_v1", "lpd_v2", "lpd_v3"]
DOORBELL_PRO_2_KINDS = ["lpd_v4"]
DOORBELL_ELITE_KINDS = ["jbox_v1"]
DOORBELL_WIRED_KINDS = ["doorbell_graham_cracker"]
DOORBELL_BATTERY_KINDS = ["df_doorbell_clownfish"]
PEEPHOLE_CAM_KINDS = ["doorbell_portal"]
DOORBELL_GEN2_KINDS = ["cocoa_doorbell", "cocoa_doorbell_v2"]

FLOODLIGHT_CAM_KINDS = ["hp_cam_v1", "floodlight_v2"]
FLOODLIGHT_CAM_PRO_KINDS = ["floodlight_pro"]
FLOODLIGHT_CAM_PLUS_KINDS = ["cocoa_floodlight"]
INDOOR_CAM_KINDS = ["stickup_cam_mini"]
INDOOR_CAM_GEN2_KINDS = ["stickup_cam_mini_v2"]
SPOTLIGHT_CAM_BATTERY_KINDS = ["stickup_cam_v4"]
SPOTLIGHT_CAM_WIRED_KINDS = ["hp_cam_v2", "spotlightw_v2"]
SPOTLIGHT_CAM_PLUS_KINDS = ["cocoa_spotlight"]
SPOTLIGHT_CAM_PRO_KINDS = ["stickup_cam_longfin"]
STICKUP_CAM_KINDS = ["stickup_cam", "stickup_cam_v3"]
STICKUP_CAM_BATTERY_KINDS = ["stickup_cam_lunar"]
STICKUP_CAM_ELITE_KINDS = ["stickup_cam_elite", "stickup_cam_wired"]
STICKUP_CAM_WIRED_KINDS = STICKUP_CAM_ELITE_KINDS  # Deprecated
STICKUP_CAM_GEN3_KINDS = ["cocoa_camera"]
BEAM_KINDS = ["beams_ct200_transformer"]

INTERCOM_KINDS = ["intercom_handset_audio"]

# error strings
MSG_BOOLEAN_REQUIRED = "Boolean value is required."
MSG_EXISTING_TYPE = f"Integer value where {DOORBELL_EXISTING_TYPE}."
MSG_GENERIC_FAIL = "Sorry.. Something went wrong..."
FILE_EXISTS = "The file {0} already exists."
MSG_VOL_OUTBOUND = "Must be within the {0}-{1}."
MSG_ALLOWED_VALUES = "Only the following values are allowed: {0}."
MSG_EXPECTED_ATTRIBUTE_NOT_FOUND = "Couldn't find expected attribute: {0}."

PUSH_ACTION_DING = "com.ring.push.HANDLE_NEW_DING"
PUSH_ACTION_MOTION = "com.ring.push.HANDLE_NEW_motion"
PUSH_ACTION_INTERCOM_UNLOCK = "com.ring.push.INTERCOM_UNLOCK_FROM_APP"

PUSH_NOTIFICATION_DING = "com.ring.pn.live-event.ding"
PUSH_NOTIFICATION_MOTION = "com.ring.pn.live-event.motion"
PUSH_NOTIFICATION_INTERCOM = "com.ring.pn.live-event.intercom"
PUSH_NOTIFICATION_INTERCOM_UNLOCK = "com.ring.pn.intercom.virtual.unlock"

PUSH_NOTIFICATION_KINDS = {
    PUSH_ACTION_DING: KIND_DING,  # legacy
    PUSH_NOTIFICATION_DING: KIND_DING,
    PUSH_ACTION_MOTION: KIND_MOTION,  # legacy
    PUSH_NOTIFICATION_MOTION: KIND_MOTION,
    PUSH_NOTIFICATION_INTERCOM: KIND_DING,
    PUSH_ACTION_INTERCOM_UNLOCK: KIND_INTERCOM_UNLOCK,
    PUSH_NOTIFICATION_INTERCOM_UNLOCK: KIND_INTERCOM_UNLOCK,
    "com.ring.push.HANDLE_NEW_SECURITY_PANEL_MODE_NONE_NOTICE": KIND_ALARM_MODE_NONE,
    "com.ring.push.HANDLE_NEW_SECURITY_PANEL_MODE_SOME_NOTICE": KIND_ALARM_MODE_SOME,
    "com.ring.push.HANDLE_NEW_USER_SOUND_SIREN": KIND_ALARM_SIREN,
    "com.ring.push.HANDLE_NEW_NON_ALARM_SIREN_SILENCED": KIND_ALARM_SILENCED,
}

POST_DATA_JSON = {
    "api_version": API_VERSION,
    "device_model": "ring-doorbell",
}

POST_DATA = {
    "api_version": API_VERSION,
    "device[os]": "android",
    "device[app_brand]": "ring",
    "device[metadata][device_model]": "KVM",
    "device[metadata][device_name]": "Python",
    "device[metadata][resolution]": "600x800",
    "device[metadata][app_version]": "1.3.806",
    "device[metadata][app_instalation_date]": "",
    "device[metadata][manufacturer]": "Qemu",
    "device[metadata][device_type]": "desktop",
    "device[metadata][architecture]": "desktop",
    "device[metadata][language]": "en",
}

PERSIST_TOKEN_DATA = {
    "api_version": API_VERSION,
    "device[metadata][device_model]": "KVM",
    "device[metadata][device_name]": "Python",
    "device[metadata][resolution]": "600x800",
    "device[metadata][app_version]": "1.3.806",
    "device[metadata][app_instalation_date]": "",
    "device[metadata][manufacturer]": "Qemu",
    "device[metadata][device_type]": "desktop",
    "device[metadata][architecture]": "x86",
    "device[metadata][language]": "en",
}

ICE_SERVERS = [
    "stun:stun.kinesisvideo.us-east-1.amazonaws.com:443",
    "stun:stun.kinesisvideo.us-east-2.amazonaws.com:443",
    "stun:stun.kinesisvideo.us-west-2.amazonaws.com:443",
    "stun:stun.l.google.com:19302",
    "stun:stun1.l.google.com:19302",
    "stun:stun2.l.google.com:19302",
    "stun:stun3.l.google.com:19302",
    "stun:stun4.l.google.com:19302",
]



# vim:sw=4:ts=4:et
# Many thanks to @troopermax <https://github.com/troopermax>
"""Python Ring command line interface."""

from __future__ import annotations

import asyncio
import functools
import getpass
import json
import logging
import select
import sys
from contextlib import asynccontextmanager
from datetime import datetime
from pathlib import Path, PurePath
from typing import Sequence, cast, TypeVar, NoReturn

import asyncclick as click

from ring_doorbell import (
    Auth,
    AuthenticationError,
    Requires2FAError,
    Ring,
    RingDoorBell,
    RingEvent,
    RingGeneric,
    RingStickUpCam,
    RingOther,
    RingCapability,
)
from ring_doorbell.const import (
    CLI_TOKEN_FILE,
    GCM_TOKEN_FILE,
    PACKAGE_NAME,
    USER_AGENT,
    DOORBELL_EXISTING_TYPE,
)


def _header() -> None:
    _bar()
    echo("Ring CLI")


def _bar() -> None:
    echo("---------------------------------")


def error(msg: str) -> NoReturn:
    """Print an error and exit."""
    echo(msg)
    sys.exit(1)


click.anyio_backend = "asyncio"  # type: ignore[attr-defined]

pass_ring = click.make_pass_decorator(Ring)
pass_doorbell = click.make_pass_decorator(RingDoorBell)

cache_file = Path(CLI_TOKEN_FILE)
gcm_cache_file = Path(GCM_TOKEN_FILE)


def CatchAllExceptions(cls):
    """Capture all exceptions and prints them nicely.

    Idea from https://stackoverflow.com/a/44347763 and
    https://stackoverflow.com/questions/52213375
    """

    def _handle_exception(debug, exc):
        if isinstance(exc, click.ClickException):
            raise
        # Handle exit request from click.
        if isinstance(exc, click.exceptions.Exit):
            sys.exit(exc.exit_code)
        echo(f"Raised error: {exc}")
        if debug:
            raise
        echo("Run with --debug enabled to see stacktrace")
        sys.exit(1)

    class _CommandCls(cls):
        _debug = False

        async def make_context(self, info_name, args, parent=None, **extra):
            self._debug = any(
                [arg for arg in args if arg in ["--debug", "-d", "--verbose", "-v"]]
            )
            try:
                return await super().make_context(
                    info_name, args, parent=parent, **extra
                )
            except Exception as exc:
                _handle_exception(self._debug, exc)

        async def invoke(self, ctx):
            try:
                return await super().invoke(ctx)
            except asyncio.CancelledError:
                pass
            except KeyboardInterrupt:
                echo("Cli interrupted with keyboard interrupt")
            except Exception as exc:
                _handle_exception(self._debug, exc)

    return _CommandCls


class MutuallyExclusiveOption(click.Option):
    """Prevents incompatable options being supplied, i.e. on and off."""

    def __init__(self, *args, **kwargs) -> None:
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        _help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(self.mutually_exclusive)
            kwargs["help"] = _help + (
                " NOTE: This argument is mutually exclusive with "
                " arguments: [" + ex_str + "]."
            )
        super().__init__(*args, **kwargs)

    async def handle_parse_result(self, ctx, opts, args):
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            msg = (
                "Illegal usage: `{}` is mutually exclusive with "
                "arguments `{}`.".format(self.name, ", ".join(self.mutually_exclusive))
            )
            raise click.UsageError(msg)

        return await super().handle_parse_result(ctx, opts, args)


echo = click.echo


def token_updated(token) -> None:
    """Writes token to file."""
    cache_file.write_text(json.dumps(token), encoding="utf-8")


def _format_filename(device_name, event):
    if not isinstance(event, dict):
        return None

    answered_status = "answered" if event["answered"] else "not_answered"

    filename = "{}_{}_{}_{}_{}".format(
        device_name, event["created_at"], event["kind"], answered_status, event["id"]
    )

    return filename.replace(" ", "_").replace(":", ".") + ".mp4"


async def _do_auth(username, password, user_agent=USER_AGENT):
    if not username:
        username = input("Username: ")

    if not password:
        password = getpass.getpass("Password: ")

    auth = Auth(user_agent, None, token_updated)
    try:
        await auth.async_fetch_token(username, password)
        return auth
    except Requires2FAError:
        await auth.async_fetch_token(username, password, input("2FA Code: "))
        return auth


async def _get_ring(username, password, do_update_data, user_agent=USER_AGENT):
    # connect to Ring account
    global cache_file, gcm_cache_file
    if user_agent != USER_AGENT:
        cache_file = Path(user_agent + ".token.cache")
        gcm_cache_file = Path(user_agent + ".gcm_token.cache")
    if cache_file.is_file():
        auth = Auth(
            user_agent,
            json.loads(cache_file.read_text(encoding="utf-8")),
            token_updated,
        )
        ring = Ring(auth)
        do_method = (
            ring.async_update_data if do_update_data else ring.async_create_session
        )
        try:
            await do_method()
        except AuthenticationError:
            auth = await _do_auth(username, password)
            ring = Ring(auth)
            do_method = (
                ring.async_update_data if do_update_data else ring.async_create_session
            )
            await do_method()
    else:
        auth = await _do_auth(username, password, user_agent=user_agent)
        ring = Ring(auth)
        do_method = (
            ring.async_update_data if do_update_data else ring.async_create_session
        )
        await do_method()

    return ring


_T = TypeVar("_T")


def _get_device(
    ring: Ring,
    device_families: list[str],
    device_type: type[_T],
    device_name: str | None = None,
    *,
    device_description: str | None = None,
) -> _T:
    description = (
        device_description if device_description else " or ".join(device_families)
    )
    if not device_name:
        dev_dict: dict[int, RingGeneric] = {}
        devices = ring.devices()
        for device_family in device_families:
            for dev in devices[device_family]:
                dev_dict[dev.device_api_id] = dev
        devs = list(dev_dict.values())
        found = len(dev_dict)
        if found == 1:
            return cast(_T, devs[0])
        elif found == 0:
            error(f"No {description} found")
        else:
            error(
                f"There are {found} {description}s, you need to pass the --device-name option."
            )
    elif device := ring.get_device_by_name(device_name):
        if device.family in device_families and isinstance(device, device_type):
            return device
        else:
            error(f"{device_name} is not a {description}")
    else:
        error(f"Cannot find {description} with name {device_name}")


@click.group(
    invoke_without_command=True,
    cls=CatchAllExceptions(click.Group),
)
@click.version_option(package_name="ring_doorbell")
@click.option(
    "--username",
    default=None,
    required=False,
    envvar="RING_USERNAME",
    help="Username for Ring account.",
)
@click.option(
    "--password",
    default=None,
    required=False,
    envvar="RING_PASSWORD",
    help="Password for Ring account",
)
@click.option("-d", "--debug", default=False, is_flag=True)
@click.option(
    "--user-agent",
    default=USER_AGENT,
    required=False,
    envvar="RING_USER_AGENT",
    help="User agent to send to ring",
)
@click.pass_context
async def cli(ctx, username, password, debug, user_agent):
    """Command line function."""
    # no need to perform any checks if we are just displaying the help
    if "--help" in sys.argv:
        # Context object is required to avoid crashing on sub-groups
        ctx.obj = Ring(None)
        return
    _header()

    logging.basicConfig()
    log_level = logging.DEBUG if debug else logging.INFO
    logger = logging.getLogger(PACKAGE_NAME)
    logger.setLevel(log_level)

    logger = logging.getLogger("firebase_messaging")
    logger.setLevel(log_level)

    no_update_commands = ["listen"]
    no_update = ctx.invoked_subcommand in no_update_commands

    @asynccontextmanager
    async def async_wrapped_ring(ring: Ring):
        try:
            yield ring
        finally:
            await ring.auth.async_close()

    ring = await _get_ring(username, password, not no_update, user_agent)
    # wrapped ring will ensure async_close is called when cli is finished
    ctx.obj = await ctx.with_async_resource(async_wrapped_ring(ring))

    if ctx.invoked_subcommand is None:
        return await ctx.invoke(show)
    return None


@cli.command(name="list")
@pass_ring
async def list_command(ring: Ring) -> None:
    """List ring devices."""
    devices = ring.devices()
    device: RingGeneric | None = None
    for device in devices.doorbots:
        echo(device)
    for device in devices.authorized_doorbots:
        echo(device)
    for device in devices.chimes:
        echo(device)
    for device in devices.stickup_cams:
        echo(device)
    for device in devices.other:
        echo(device)


@cli.command()
@pass_ring
@click.pass_context
@click.option(
    "--on",
    "turn_on",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    default=None,
    required=False,
    mutually_exclusive=["--off"],
)
@click.option(
    "--off",
    "turn_off",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    default=None,
    required=False,
    mutually_exclusive=["--on"],
)
@click.option(
    "--device-name",
    "-dn",
    required=True,
    default=None,
    help="Name of the ring device",
)
async def motion_detection(ctx, ring: Ring, device_name, turn_on, turn_off):
    """Get and change the motion detecton status of a device."""
    device = ring.get_device_by_name(device_name)

    if not device:
        echo(
            f"No device with name {device_name} found."
            + " List of found device names (kind) is:"
        )
        return await ctx.invoke(list_command)
    if not device.has_capability("motion_detection"):
        echo(f"{device!s} is not capable of motion detection")
        return None
    device = cast(RingDoorBell, device)
    state = "on" if device.motion_detection else "off"
    if not turn_on and not turn_off:
        echo(f"{device!s} has motion detection {state}")
        return None
    is_on = device.motion_detection
    if (turn_on and is_on) or (turn_off and not is_on):
        echo(f"{device!s} already has motion detection {state}")
        return None

    await device.async_set_motion_detection(turn_on if turn_on else False)
    await ring.async_update_devices()
    state = "on" if device.motion_detection else "off"
    echo(f"{device!s} motion detection set to {state}")
    return None


@cli.command()
@pass_ring
@click.pass_context
@click.argument("enable", type=click.BOOL, default=None, required=False)
@click.option(
    "--device-name",
    "-dn",
    required=True,
    default=None,
    help="Name of the ring device",
)
async def light(ctx, ring: Ring, device_name, enable):
    """Get and change the light state of a device."""
    device = ring.get_device_by_name(device_name)

    if not device:
        echo(
            f"No device with name {device_name} found."
            + " List of found device names (kind) is:"
        )
        return await ctx.invoke(list_command)
    if not device.has_capability(RingCapability.LIGHT):
        echo(f"{device!s} does not have a light")
        return None
    device = cast(RingStickUpCam, device)
    state = "on" if device.light else "off"
    if enable is None:
        echo(f"{device!s} light is {state}")
        return None
    is_on = device.light
    if (enable and is_on) or (not enable and not is_on):
        echo(f"{device!s} light is already {state}")
        return None

    await device.async_set_light(enable)
    await ring.async_update_devices()
    state = "on" if device.light else "off"
    echo(f"{device!s} light set to {state}")
    return None


@cli.command()
@click.option(
    "--device-name",
    "-dn",
    required=False,
    default=None,
    help="Name of device, if ommited shows all devices",
)
@pass_ring
@click.pass_context
async def show(ctx, ring: Ring, device_name):
    """Display ring devices."""
    devices: Sequence[RingGeneric] | None = None

    if device_name and (device := ring.get_device_by_name(device_name)):
        devices = [device]
    elif device_name:
        echo(
            f"No device with name {device_name} found. "
            + "List of found device names (kind) is:"
        )
        return await ctx.invoke(list_command)
    else:
        devices = ring.get_device_list()

    for dev in devices:
        await dev.async_update_health_data()
        echo("Name:       %s" % dev.name)
        echo("Family:     %s" % dev.family)
        echo("ID:         %s" % dev.id)
        echo("Timezone:   %s" % dev.timezone)
        echo("Wifi Name:  %s" % dev.wifi_name)
        echo("Wifi RSSI:  %s" % dev.wifi_signal_strength)
        echo()
    return None


@cli.command(name="devices")
@click.option(
    "--device-name",
    "-dn",
    required=False,
    default=None,
    help="Name of device, if ommited shows all devices",
)
@click.option(
    "--json",
    "json_flag",
    required=False,
    is_flag=True,
    help="Output raw json",
)
@pass_ring
@click.pass_context
async def devices_command(ctx, ring: Ring, device_name, json_flag):
    """Get device information."""
    if not json_flag:
        echo(
            "(Pretty format coming soon, if you want json consistently "
            + "from this command provide the --json flag)"
        )
    device_json = None
    if device_name and (device := ring.get_device_by_name(device_name)):
        device_json = ring.devices_data[device.family][device.id]
    elif device_name:
        echo(
            f"No device with name {device_name} found. "
            + "List of found device names (kind) is:"
        )
        return await ctx.invoke(list_command)

    if device_json:
        echo(json.dumps(device_json, indent=2))
        return None
    else:
        for device_type in ring.devices_data:
            for device_api_id in ring.devices_data[device_type]:
                echo(
                    json.dumps(ring.devices_data[device_type][device_api_id], indent=2)
                )
        return None


@cli.command()
@click.option(
    "--json",
    "json_flag",
    required=False,
    is_flag=True,
    help="Output raw json",
)
@pass_ring
async def dings(ring: Ring, json_flag) -> None:
    """Get dings information."""
    if not json_flag:
        echo(
            "(Pretty format coming soon, if you want json consistently "
            + "from this command provide the --json flag)"
        )
    echo(json.dumps(ring.dings_data, indent=2))


@cli.command()
@click.option(
    "--json",
    "json_flag",
    required=False,
    is_flag=True,
    help="Output raw json",
)
@pass_ring
async def groups(ring: Ring, json_flag) -> None:
    """Get group information."""
    if not json_flag:
        echo(
            "(Pretty format coming soon, if you want json consistently "
            + "from this command provide the --json flag)"
        )
    if not ring.groups_data:
        echo("No ring device groups setup")
    else:
        for light_group in ring.groups().values():
            await light_group.async_update()
            echo(json.dumps(light_group._attrs, indent=2))
            echo(json.dumps(light_group._health_attrs, indent=2))


@cli.command()
@click.option(
    "--url",
    required=True,
    type=str,
    help="Url to query, i.e. /clients_api/dings/active",
)
@pass_ring
async def raw_query(ring: Ring, url) -> None:
    """Directly query a url and return json result."""
    resp = await ring.async_query(url)
    data = resp.json()
    echo(json.dumps(data, indent=2))


@cli.command(name="history")
@click.option(
    "--device-name",
    "-dn",
    required=False,
    default=None,
    help="Name of device, if ommited shows all devices",
)
@click.option(
    "--limit",
    required=False,
    default=5,
    help="Limit number of records to return",
)
@click.option(
    "--kind",
    required=False,
    default=None,
    type=click.Choice(["ding", "motion", "on_demand"], case_sensitive=False),
    help="Get devices",
)
@click.option(
    "--json",
    "json_flag",
    required=False,
    is_flag=True,
    help="Output raw json",
)
@pass_ring
@click.pass_context
async def history_command(ctx, ring: Ring, device_name, kind, limit, json_flag):
    """Print raw json."""
    if not json_flag:
        echo(
            "(Pretty format coming soon, if you want json consistently "
            + "from this command provide the --json flag)"
        )
    device = ring.get_device_by_name(device_name)
    if not device:
        echo(
            f"No device with name {device_name} found. "
            + "List of found device names (kind) is:"
        )
        return await ctx.invoke(list_command)

    history = await device.async_history(limit=limit, kind=kind, convert_timezone=False)
    echo(json.dumps(history, indent=2))
    return None


@cli.command()
@click.option(
    "--count",
    required=False,
    default=False,
    is_flag=True,
    help="Count the number of videos on your Ring account",
)
@click.option(
    "--download-all",
    required=False,
    default=False,
    is_flag=True,
    help="Download all videos on your Ring account",
)
@click.option(
    "--download",
    required=False,
    default=False,
    is_flag=True,
    help="Download videos on your Ring account up to the max-count option",
)
@click.option(
    "--max-count",
    required=False,
    default=300,
    help="Maximum count of videos to count or download from your Ring account",
)
@click.option(
    "--download-to",
    required=False,
    default="./",
    help="Download location ending with a /",
)
@click.option(
    "--device-name",
    "-dn",
    default=None,
    required=False,
    help="Name of the ring device, if omitted uses the first device returned",
)
@pass_ring
@click.pass_context
async def videos(
    ctx, ring: Ring, count, download, download_all, max_count, download_to, device_name
):
    """Interact with ring videos."""
    device = None
    if device_name and not (device := ring.get_video_device_by_name(device_name)):
        echo(
            f"No device with name {device_name} found. "
            + "List of found device names (kind) is:"
        )
        return await ctx.invoke(list_command)
    if device and not device.has_capability("video"):
        echo(f"Device {device.name} is not a video device")
        return None
    # return the first device is implemented to be consistent with previous cli version
    if not device:
        if video_devices := ring.video_devices():
            device = video_devices[0]
        else:
            echo(
                "No video devices found. "
                + "List of found device names (with device kind) is:"
            )
            return await ctx.invoke(list_command)

    if not device:  # Make mypy happy
        return None
    if (
        not count
        and not download
        and not download_all
        and (last_recording_id := await device.async_get_last_recording_id())
        and (url := await device.async_recording_url(last_recording_id))
    ):
        echo("Last recording url is: " + url)
        return None

    events = None
    if download_all:
        download = True
        max_count = -1

    async def _get_events(device, max_count):
        limit = 100 if max_count == -1 else min(100, max_count)
        events = []
        history = await device.async_history(limit=limit)
        while len(history) > 0:
            events += history
            if (len(events) >= max_count and max_count != -1) or len(history) < limit:
                break
            history = await device.async_history(
                older_than=history[-1]["id"], limit=limit
            )
        return events

    if count:
        echo(
            f"\tCounting videos linked on your Ring account for {device.name}.\n"
            + "\tThis may take some time....\n"
        )

        events = await _get_events(device, max_count)

        motion = len([m["kind"] for m in events if m["kind"] == "motion"])
        ding = len([m["kind"] for m in events if m["kind"] == "ding"])
        on_demand = len([m["kind"] for m in events if m["kind"] == "on_demand"])

        echo(f"\tTotal videos: {len(events)}")
        echo(f"\tDing triggered: {ding}")
        echo(f"\tMotion triggered: {motion}")
        echo(f"\tOn-Demand triggered: {on_demand}")

    if download:
        if events is None:
            echo(
                "\tGetting videos linked on your Ring account.\n"
                "\tThis may take some time....\n"
            )
            events = await _get_events(device, max_count)

        echo(
            f"\tDownloading {len(events)} videos linked on your Ring account.\n"
            "\tThis may take some time....\n"
        )
        for counter, event in enumerate(events):
            filename = str(PurePath(download_to, _format_filename(device.name, event)))
            echo(f"\t{counter}/{len(events)} Downloading {filename}")

            await device.async_recording_download(
                event["id"], filename=filename, override=False
            )
        return None
    return None


@cli.group(invoke_without_command=True)
@pass_ring
@click.pass_context
@click.option(
    "--device-name",
    "-dn",
    required=False,
    default=None,
    help="Name of the ring device",
)
async def in_home_chime(ctx, ring: Ring, device_name):
    """View and manage the Doorbell in-home chime. To see the current in-home chime status of a device, only pass the device name."""
    if "--help" in sys.argv:
        return
    device = _get_device(
        ring,
        ["authorized_doorbots", "doorbots"],
        RingDoorBell,
        device_name,
        device_description="doorbell",
    )

    if ctx.invoked_subcommand is None:
        echo("Name:       %s" % device.name)
        echo("ID:         %s" % device.id)
        echo("Type:       %s" % device.existing_doorbell_type)
        echo("Enabled:    %s" % device.existing_doorbell_type_enabled)
        echo("Duration:   %s" % device.existing_doorbell_type_duration)
        return None

    ctx.obj = device


@in_home_chime.command(name="type")
@pass_doorbell
@click.pass_context
@click.argument(
    "new_type",
    type=click.Choice(list(DOORBELL_EXISTING_TYPE.values()), case_sensitive=False),
    default=None,
    required=False,
)
async def in_home_chime_type(ctx, device: RingDoorBell, new_type):
    """Get/set the type of In-home chime."""
    if new_type is None:
        echo(device.existing_doorbell_type)
        return

    if device.family == "authorized_doorbots":
        exit(
            f"{device.name} is a shared device and you do not have permission to update this value"
        )
    new_type_int = next(k for k, v in DOORBELL_EXISTING_TYPE.items() if v == new_type)

    await device.async_set_existing_doorbell_type(new_type_int)
    echo(f"{device.name}'s in-home chime type has been set to {new_type}")


@in_home_chime.command()
@pass_doorbell
@click.pass_context
@click.argument("enable", type=click.BOOL, default=None, required=False)
async def enabled(ctx, device: RingDoorBell, enable: bool | None):
    """Gets/sets the in-home chime enabled status.

    ENABLE: 1/0, true/false, t/f, yes/no, y/n, and on/off
    """
    if enable is None:
        echo(device.existing_doorbell_type_enabled)
        return

    if device.family == "authorized_doorbots":
        exit(
            f"{device.name} is a shared device and you do not have permission to update this value"
        )
    await device.async_set_existing_doorbell_type_enabled(enable)
    echo(
        f"{device.name}'s in-home chime has been {'enabled' if enable else 'disabled'}"
    )


@in_home_chime.command()
@pass_doorbell
@click.pass_context
@click.argument("duration", type=click.IntRange(0, 100), default=None, required=False)
async def duration(ctx, device: RingDoorBell, duration: int | None):
    """Gets/sets the in-home chime duration.

    DURATION: Value between 0 and 100
    """
    if duration is None:
        echo(device.existing_doorbell_type_duration)
        return

    if device.family == "authorized_doorbots":
        exit(
            f"{device.name} is a shared device and you do not have permission to update this value"
        )
    await device.async_set_existing_doorbell_type_duration(int(duration))
    echo(f"{device.name}'s in-home chime duration has been set to {duration} seconds")


async def ainput(string: str):
    loop = asyncio.get_event_loop()
    await loop.run_in_executor(None, lambda s=string: sys.stdout.write(s + " "))  # type: ignore[misc]

    return await loop.run_in_executor(None, sys.stdin.readline)


def get_now_str():
    return str(datetime.utcnow())


class _event_handler:  # pylint:disable=invalid-name
    def __init__(self, ring: Ring) -> None:
        self.ring = ring

    def on_event(self, event: RingEvent) -> None:
        msg = (
            get_now_str()
            + ": "
            + str(event)
            + " : Currently active count = "
            + str(len(self.ring.push_dings_data))
        )
        echo(msg)


@cli.command
@click.option(
    "--credentials-file",
    required=False,
    default=None,
    help=(
        "File to store push credentials, "
        + "if not provided credentials will be recreated from scratch"
    ),
)
@click.option(
    "--store-credentials/--no-store-credentials",
    default=True,
    help="Whether or not to store the push credentials, default is false",
)
@click.option(
    "--show-credentials",
    default=False,
    is_flag=True,
    help="Whether or not to show the push credentials, default is false",
)
@pass_ring
@click.pass_context
async def listen(
    ctx,
    ring,
    store_credentials,
    credentials_file,
    show_credentials,
) -> None:
    """Listen to push notification like the ones sent to your phone."""

    from ring_doorbell.listen import (  # pylint:disable=import-outside-toplevel
        RingEventListener,
    )

    def credentials_updated_callback(credentials) -> None:
        if store_credentials:
            with open(credentials_file, "w", encoding="utf-8") as f:
                json.dump(credentials, f)
        else:
            echo("New push credentials created:")
            if show_credentials:
                echo(credentials)

    if not credentials_file:
        credentials_file = gcm_cache_file
    else:
        credentials_file = Path(credentials_file)

    credentials = None
    if store_credentials and credentials_file.is_file():
        # already registered, load previous credentials
        with open(credentials_file, encoding="utf-8") as f:
            credentials = json.load(f)

    event_listener = RingEventListener(ring, credentials, credentials_updated_callback)
    await event_listener.start()
    event_listener.add_notification_callback(_event_handler(ring).on_event)

    await ainput("Listening, press enter to cancel\n")

    await event_listener.stop()


@cli.command
@click.option(
    "--device-name",
    required=False,
    default=None,
    help=("Name of the intercom if there are more than one."),
)
@pass_ring
@click.pass_context
async def open_door(ctx, ring: Ring, device_name: str | None) -> None:
    """Open the door of a intercom device."""
    device = _get_device(
        ring,
        ["intercoms", "other"],
        RingOther,
        device_name,
        device_description="intercom",
    )
    await device.async_open_door()
    echo(f"{device.name} opened")


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



"""Test module which runs the first example in the README."""

import getpass
import json
from pathlib import Path

from ring_doorbell import Auth, AuthenticationError, Requires2FAError, Ring

user_agent = "YourProjectName-1.0"  # Change this
cache_file = Path(user_agent + ".token.cache")


def token_updated(token) -> None:
    cache_file.write_text(json.dumps(token))


def otp_callback():
    return input("2FA code: ")


def do_auth():
    username = input("Username: ")
    password = getpass.getpass("Password: ")
    auth = Auth(user_agent, None, token_updated)
    try:
        auth.fetch_token(username, password)
    except Requires2FAError:
        auth.fetch_token(username, password, otp_callback())
    return auth


def main() -> None:
    if cache_file.is_file():  # auth token is cached
        auth = Auth(user_agent, json.loads(cache_file.read_text()), token_updated)
        ring = Ring(auth)
        try:
            ring.create_session()  # auth token still valid
        except AuthenticationError:  # auth token has expired
            auth = do_auth()
    else:
        auth = do_auth()  # Get new auth token
        ring = Ring(auth)

    ring.update_data()

    print(ring.devices())


if __name__ == "__main__":
    main()



"""Python Ring Doorbell RTC Stream handler.

This module is currently experimental and requires a webrtc enabled client
to function.
"""

from __future__ import annotations

import asyncio
import contextlib
import logging
import ssl
import time
import uuid
from dataclasses import dataclass
from json import dumps as json_dumps
from json import loads as json_loads
from typing import TYPE_CHECKING, Any, Callable

from async_timeout import timeout as asyncio_timeout
from typing_extensions import TypeAlias
from websockets.asyncio.client import connect

from ring_doorbell.const import (
    APP_API_URI,
    RTC_STREAMING_TICKET_ENDPOINT,
    RTC_STREAMING_WEB_SOCKET_ENDPOINT,
)
from ring_doorbell.exceptions import RingError

if TYPE_CHECKING:
    from collections.abc import Coroutine

    from websockets.asyncio.client import ClientConnection

    from .ring import Ring

_LOGGER = logging.getLogger(__name__)

SDP_ANSWER_TIMEOUT = 1


@dataclass
class RingWebRtcMessage:
    """Class for async on_message callback."""

    answer: str | None = None
    candidate: str | None = None
    sdp_m_line_index: int | None = None
    error_code: str | None = None
    error_message: str | None = None
    session_id: str | None = None


RingWebRtcMessageCallback: TypeAlias = Callable[[RingWebRtcMessage], None]


class RingWebRtcStream:
    """Class to handle a Web RTC Stream."""

    PING_TIME_SECONDS = 5

    def __init__(
        self,
        ring: Ring,
        device_api_id: int,
        *,
        keep_alive_timeout: int | None = 30,
        on_message_callback: RingWebRtcMessageCallback | None = None,
        on_close_callback: Callable[[], Coroutine[Any, Any, None]] | None = None,
    ) -> None:
        """Initialise the class."""
        self._ring = ring
        self.device_api_id = device_api_id
        self.sdp: str | None = None
        self.websocket: ClientConnection | None = None
        self.is_alive = True
        self.ping_task: asyncio.Task | None = None
        self.read_task: asyncio.Task | None = None
        self._close_task: asyncio.Task | None = None
        self.ice_candidates: dict[int, list[str]] = {0: [], 1: [], 2: [], 3: []}
        self.collect_ice_candidates = False
        self.ssl_context: ssl.SSLContext | None = None
        self._sdp_answer_event = asyncio.Event()
        self._keep_alive_timeout = keep_alive_timeout
        self._last_keep_alive: float | None = None
        self._on_close_callback = on_close_callback
        self._on_message_callback = on_message_callback
        self.session_id: str | None = None
        self._offered_event = asyncio.Event()

    @staticmethod
    def get_sdp_session_id(sdp_offer: str) -> str | None:
        """Return the sdp session id from the offer."""
        try:
            lines = sdp_offer.split("\n")
            for line in lines:
                if line[0] == "o":
                    origin = line.split("=")[1]
                    break
            return origin.split(" ")[1]
        except Exception:
            _LOGGER.exception("Error getting session id from offer: %s", sdp_offer)
            return None

    async def generate(self, sdp_offer: str) -> str | None:
        """Generate the RTC stream."""
        await self._generate(sdp_offer)

        if self._on_message_callback:
            return None

        if self.collect_ice_candidates:
            _LOGGER.debug(
                "Waiting %s seconds for ice candidates",
                SDP_ANSWER_TIMEOUT,
            )
            await asyncio.sleep(SDP_ANSWER_TIMEOUT)
            self.insert_ice_candidates()
        else:
            async with asyncio_timeout(SDP_ANSWER_TIMEOUT):
                await self._sdp_answer_event.wait()

        if not self.sdp:
            exmsg = "Unable to generate RTC stream in time"
            await self.close()
            raise RingError(exmsg)
        _LOGGER.debug("Returning SDP answer: %s", self.sdp)
        return self.sdp

    async def _generate(self, sdp_offer: str) -> None:
        """Generate the RTC stream."""
        try:
            _LOGGER.debug("Generating stream with sdp offer: %s", sdp_offer)
            req = await self._ring.async_query(
                RTC_STREAMING_TICKET_ENDPOINT,
                method="POST",
                base_uri=APP_API_URI,
            )
            ticket = req.json()["ticket"]

            _LOGGER.debug(
                "Received RTC streaming ticket %s from endpoint, creating websocket",
                ticket,
            )
            ws_uri = RTC_STREAMING_WEB_SOCKET_ENDPOINT.format(uuid.uuid4(), ticket)
            loop = asyncio.get_running_loop()

            if not self.ssl_context:
                # create_default_context() blocks the event loop
                self.ssl_context = await loop.run_in_executor(
                    None, ssl.create_default_context
                )
            self.websocket = await connect(
                ws_uri,
                user_agent_header="android:com.ringapp",
                ssl=self.ssl_context,
            )

            self.dialog_id = str(uuid.uuid4())
            offer_msg = {
                "method": "live_view",
                "dialog_id": self.dialog_id,
                "body": {
                    "doorbot_id": self.device_api_id,
                    "stream_options": {"audio_enabled": True, "video_enabled": True},
                    "sdp": sdp_offer,
                    "type": "offer",
                },
            }
            _LOGGER.debug(
                "Connected to RTC streaming websocket, sending live_view offer msg: %s",
                offer_msg,
            )
            _LOGGER.debug("Starting reader task")
            self.read_task = asyncio.create_task(self.reader())

            await self.websocket.send(json_dumps(offer_msg))

            self._offered_event.set()

        except Exception as ex:
            exmsg = "Error generating RTC stream"
            raise RingError(exmsg, ex) from ex

    async def _activate(self) -> None:
        if TYPE_CHECKING:
            assert self.websocket
        activate_msg = self.get_session_message("activate_session", {})
        _LOGGER.debug("Sending activate_session message: %s", activate_msg)
        await self.websocket.send(json_dumps(activate_msg))

        self._last_keep_alive = time.time()
        self.ping_task = asyncio.create_task(self.pinger())

    async def on_ice_candidate(self, candidate: str, m_line_index: int) -> None:
        """Send an ICE candidate."""
        async with asyncio_timeout(10):
            await self._offered_event.wait()

        assert self.websocket  # noqa: S101

        body = {
            "doorbot_id": self.device_api_id,
            "ice": candidate,
            "mlineindex": m_line_index,
        }
        # Set the session if it's been created.
        if self.session_id:
            body["session_id"] = self.session_id
        ice_msg = {
            "method": "ice",
            "dialog_id": self.dialog_id,
            "body": body,
        }
        _LOGGER.debug(
            "Sending ice candidate for mlineindex %s: %s", m_line_index, candidate
        )
        await self.websocket.send(json_dumps(ice_msg))

    async def keep_alive(self) -> None:
        """Keep alive the rtc stream."""
        self._last_keep_alive = time.time()

    def get_session_message(self, method: str, body: dict[str, Any]) -> dict[str, Any]:
        """Get a message to send to the session."""
        return {
            "method": method,
            "dialog_id": self.dialog_id,
            "body": {
                **body,
                "doorbot_id": self.device_api_id,
                "session_id": self.session_id,
            },
        }

    async def reader(self) -> None:
        """Read messages from the websocket."""
        if TYPE_CHECKING:
            assert self.websocket
        async for message in self.websocket:
            if TYPE_CHECKING:
                assert isinstance(message, str)
            await self.handle_message(message)

    async def pinger(self) -> None:
        """Ping to keep the session alive."""
        if TYPE_CHECKING:
            assert self.websocket
            assert self._last_keep_alive

        while self.is_alive and (
            self._keep_alive_timeout is None
            or (time.time() - self._last_keep_alive) <= self._keep_alive_timeout
        ):
            await asyncio.sleep(self.PING_TIME_SECONDS)
            ping = self.get_session_message("ping", {})
            await self.websocket.send(json_dumps(ping))

    def handle_ice_message(self, message: dict) -> None:
        """Handle an ice candidate message."""
        ice_candidate = message["body"]["ice"]
        multi_line_index = message["body"]["mlineindex"]
        if self._on_message_callback:
            ice_message = RingWebRtcMessage(
                candidate=ice_candidate, sdp_m_line_index=multi_line_index
            )
            self._on_message_callback(ice_message)
        elif self.collect_ice_candidates:
            _LOGGER.debug(
                "Ice candidate received, multi_line_index: %s candidate: %s",
                multi_line_index,
                ice_candidate,
            )
            self.ice_candidates[int(multi_line_index)].append(ice_candidate)

    async def handle_answer_message(self, message: dict) -> None:
        """Handle an sdp answer message."""
        sdp = message["body"]["sdp"]
        _LOGGER.debug("SDP answer received: %s", sdp)
        self.sdp = sdp
        self._sdp_answer_event.set()

        if self._on_message_callback:
            answer_message = RingWebRtcMessage(answer=sdp)
            self._on_message_callback(answer_message)

        await self._activate()

    async def handle_close_message(self, message: dict) -> None:
        """Handle an sdp answer message."""
        reason = message["body"]["reason"]
        reason_code = reason["code"]
        reason_message = reason["text"]
        _LOGGER.debug("Close message received: %s", str(reason))
        self.is_alive = False
        await self._close(closed_by_self=True)
        if self._on_message_callback:
            error_message = RingWebRtcMessage(
                error_code=reason_code, error_message=reason_message
            )
            self._on_message_callback(error_message)

    def insert_ice_candidates(self) -> None:
        """Insert an ice candidate into the sdp answer.

        In 2023 the ring api did not return the ice servers in the sdp answer
        and they had to be added as they were received on the web socket.
        As of Sep 2024 they are coming back with the initial sdp answer.
        """
        if TYPE_CHECKING:
            assert self.sdp
        _LOGGER.debug("Inserting ICE candidates into sdp answer")
        self.collect_ice_candidates = False
        for line_index, candidates in self.ice_candidates.items():
            if not candidates:
                continue
            candidates_dict = {
                int(candidate[10:12]): f"a={candidate}\r\n" for candidate in candidates
            }
            candidates_text = ("").join(dict(sorted(candidates_dict.items())).values())
            multi_text = f"a=mid:{line_index}"
            self.sdp = self.sdp.replace(multi_text, candidates_text + multi_text)

    def sync_close(self) -> None:
        """Close a WebRTC session."""
        if self.is_alive and not self._close_task:
            self._close_task = asyncio.create_task(self.close())

    async def close(self) -> None:
        """Close the rtc stream."""
        _LOGGER.debug("Closing the RTC Stream")
        await self._close(closed_by_self=False)
        self._close_task = None

    async def _close(self, *, closed_by_self: bool) -> None:
        """Close the stream."""
        self.session_id = None
        if closed_by_self and (close_cb := self._on_close_callback):
            self._on_close_callback = None
            await close_cb()
        self.is_alive = False
        if ping_task := self.ping_task:
            self.ping_task = None
            if not ping_task.done():
                ping_task.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await ping_task
        if websocket := self.websocket:
            self.websocket = None
            await websocket.close()
        if read_task := self.read_task:
            self.read_task = None
            if not read_task.done():
                await read_task

    async def handle_message(self, message_str: str) -> None:  # noqa: C901, PLR0912
        """Handle a message from the web socket."""
        if TYPE_CHECKING:
            assert self.websocket
        message = json_loads(message_str)
        method = message["method"]
        if method == "ice":
            self.handle_ice_message(message)
        elif method == "sdp":
            await self.handle_answer_message(message)
        elif method == "notification":
            text = message["body"]["text"]
            if text == "camera_connected":
                _LOGGER.debug("Notification received: %s", text)
                camera_options = self.get_session_message(
                    "camera_options", {"stealth_mode": False}
                )
                _LOGGER.debug("Sending camera options: %s", camera_options)
                await self.websocket.send(json_dumps(camera_options))
            else:
                _LOGGER.debug("Received notification: %s", message)
        elif method == "session_created":
            self.session_id = message["body"]["session_id"]
            if TYPE_CHECKING:
                assert self.session_id
            _LOGGER.debug(
                "Session created: %s___%s", self.session_id[:16], self.session_id[-16:]
            )
        elif method == "close":
            await self.handle_close_message(message)
        elif method == "pong":
            _LOGGER.debug("Pong message received")
        elif method == "camera_started":
            _LOGGER.debug("camera_started message received")
        elif method == "camera_options":
            _LOGGER.debug("camera_options message received: %s", message["body"])
        else:
            _LOGGER.debug("Unknown message received with method: %s", method)



# vim:sw=4:ts=4:et:
"""Python Ring Doorbell wrapper."""

from __future__ import annotations

import logging
from typing import ClassVar

from ring_doorbell.const import (
    FLOODLIGHT_CAM_KINDS,
    FLOODLIGHT_CAM_PLUS_KINDS,
    FLOODLIGHT_CAM_PRO_KINDS,
    INDOOR_CAM_GEN2_KINDS,
    INDOOR_CAM_KINDS,
    LIGHTS_ENDPOINT,
    MSG_ALLOWED_VALUES,
    MSG_VOL_OUTBOUND,
    SIREN_DURATION_MAX,
    SIREN_DURATION_MIN,
    SIREN_ENDPOINT,
    SPOTLIGHT_CAM_BATTERY_KINDS,
    SPOTLIGHT_CAM_PLUS_KINDS,
    SPOTLIGHT_CAM_PRO_KINDS,
    SPOTLIGHT_CAM_WIRED_KINDS,
    STICKUP_CAM_BATTERY_KINDS,
    STICKUP_CAM_ELITE_KINDS,
    STICKUP_CAM_GEN3_KINDS,
    STICKUP_CAM_KINDS,
    RingCapability,
)
from ring_doorbell.doorbot import RingDoorBell
from ring_doorbell.exceptions import RingError

_LOGGER = logging.getLogger(__name__)


class RingStickUpCam(RingDoorBell):
    """Implementation for RingStickUpCam."""

    @property
    def family(self) -> str:
        """Return Ring device family type."""
        return "stickup_cams"

    @property
    def model(self) -> str:  # noqa: C901, PLR0911, PLR0912
        """Return Ring device model name."""
        if self.kind in FLOODLIGHT_CAM_KINDS:
            return "Floodlight Cam"
        if self.kind in FLOODLIGHT_CAM_PRO_KINDS:
            return "Floodlight Cam Pro"
        if self.kind in FLOODLIGHT_CAM_PLUS_KINDS:
            return "Floodlight Cam Plus"
        if self.kind in INDOOR_CAM_KINDS:
            return "Indoor Cam"
        if self.kind in INDOOR_CAM_GEN2_KINDS:
            return "Indoor Cam (2nd Gen)"
        if self.kind in SPOTLIGHT_CAM_BATTERY_KINDS:
            return "Spotlight Cam {}".format(
                self._attrs.get("ring_cam_setup_flow", "battery").title()
            )
        if self.kind in SPOTLIGHT_CAM_WIRED_KINDS:
            return "Spotlight Cam {}".format(
                self._attrs.get("ring_cam_setup_flow", "wired").title()
            )
        if self.kind in SPOTLIGHT_CAM_PLUS_KINDS:
            return "Spotlight Cam Plus"
        if self.kind in SPOTLIGHT_CAM_PRO_KINDS:
            return "Spotlight Cam Pro"
        if self.kind in STICKUP_CAM_KINDS:
            return "Stick Up Cam"
        if self.kind in STICKUP_CAM_BATTERY_KINDS:
            return "Stick Up Cam Battery"
        if self.kind in STICKUP_CAM_ELITE_KINDS:
            return "Stick Up Cam Wired"
        if self.kind in STICKUP_CAM_GEN3_KINDS:
            return "Stick Up Cam (3rd Gen)"
        _LOGGER.error("Unknown kind: %s", self.kind)
        return "Unknown Stickup Cam"

    def has_capability(self, capability: RingCapability | str) -> bool:
        """Return if device has specific capability."""
        capability = (
            capability
            if isinstance(capability, RingCapability)
            else RingCapability.from_name(capability)
        )
        if capability == RingCapability.HISTORY:
            return True
        if capability == RingCapability.BATTERY:
            return self.kind in (
                SPOTLIGHT_CAM_BATTERY_KINDS
                + STICKUP_CAM_KINDS
                + STICKUP_CAM_BATTERY_KINDS
                + STICKUP_CAM_GEN3_KINDS
            )
        if capability == RingCapability.LIGHT:
            return self.kind in (
                FLOODLIGHT_CAM_KINDS
                + FLOODLIGHT_CAM_PRO_KINDS
                + FLOODLIGHT_CAM_PLUS_KINDS
                + SPOTLIGHT_CAM_BATTERY_KINDS
                + SPOTLIGHT_CAM_WIRED_KINDS
                + SPOTLIGHT_CAM_PLUS_KINDS
                + SPOTLIGHT_CAM_PRO_KINDS
            )
        if capability == RingCapability.SIREN:
            return self.kind in (
                FLOODLIGHT_CAM_KINDS
                + FLOODLIGHT_CAM_PRO_KINDS
                + FLOODLIGHT_CAM_PLUS_KINDS
                + INDOOR_CAM_KINDS
                + INDOOR_CAM_GEN2_KINDS
                + SPOTLIGHT_CAM_BATTERY_KINDS
                + SPOTLIGHT_CAM_WIRED_KINDS
                + SPOTLIGHT_CAM_PLUS_KINDS
                + SPOTLIGHT_CAM_PRO_KINDS
                + STICKUP_CAM_BATTERY_KINDS
                + STICKUP_CAM_ELITE_KINDS
                + STICKUP_CAM_GEN3_KINDS
            )
        if capability in [RingCapability.MOTION_DETECTION, RingCapability.VIDEO]:
            return self.kind in (
                FLOODLIGHT_CAM_KINDS
                + FLOODLIGHT_CAM_PRO_KINDS
                + FLOODLIGHT_CAM_PLUS_KINDS
                + INDOOR_CAM_KINDS
                + INDOOR_CAM_GEN2_KINDS
                + SPOTLIGHT_CAM_BATTERY_KINDS
                + SPOTLIGHT_CAM_WIRED_KINDS
                + SPOTLIGHT_CAM_PLUS_KINDS
                + SPOTLIGHT_CAM_PRO_KINDS
                + STICKUP_CAM_KINDS
                + STICKUP_CAM_BATTERY_KINDS
                + STICKUP_CAM_ELITE_KINDS
                + STICKUP_CAM_GEN3_KINDS
            )
        return False

    @property
    def lights(self) -> str:
        """Return lights status."""
        return self._attrs.get("led_status", "")

    async def async_set_lights(self, state: str) -> None:
        """Control the lights."""
        values = ["on", "off"]
        if state not in values:
            raise RingError(MSG_ALLOWED_VALUES.format(", ".join(values)))

        url = LIGHTS_ENDPOINT.format(self.device_api_id, state)
        await self._ring.async_query(url, method="PUT")

    @property
    def light(self) -> bool:
        """Return lights status."""
        return self._attrs.get("led_status", "") == "on"

    async def async_set_light(self, value: bool) -> None:  # noqa: FBT001
        """Control the lights."""
        state = "on" if value else "off"
        url = LIGHTS_ENDPOINT.format(self.device_api_id, state)
        await self._ring.async_query(url, method="PUT")

    @property
    def siren(self) -> int:
        """Return siren status."""
        if siren_status := self._attrs.get("siren_status"):
            return siren_status.get("seconds_remaining", 0)
        return 0

    async def async_set_siren(self, duration: int) -> None:
        """Control the siren."""
        if not (
            (isinstance(duration, int))
            and (SIREN_DURATION_MIN <= duration <= SIREN_DURATION_MAX)
        ):
            raise RingError(
                MSG_VOL_OUTBOUND.format(SIREN_DURATION_MIN, SIREN_DURATION_MAX)
            )

        if duration > 0:
            state = "on"
            params = {"duration": duration}
        else:
            state = "off"
            params = {}
        url = SIREN_ENDPOINT.format(self.device_api_id, state)
        await self._ring.async_query(url, extra_params=params, method="PUT")

    DEPRECATED_API_PROPERTY_SETTERS: ClassVar = {
        *RingDoorBell.DEPRECATED_API_PROPERTY_SETTERS,
        "lights",
        "siren",
    }



"""Package for listener modules."""

from .eventlistener import RingEventListener
from .listenerconfig import RingEventListenerConfig

# can_listen used to be checkable to see if the optional listen extra installed.
# Now installed as default.
can_listen = True

__all__ = [
    "RingEventListener",
    "RingEventListenerConfig",
]



# Configuration file for the Sphinx documentation builder.  # noqa: INP001
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from __future__ import annotations

from importlib.metadata import version as _version

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "python-ring-doorbell"
copyright = "2023, Marcelo Moreira de Mello"  # noqa: A001
author = "Marcelo Moreira de Mello"
release = _version("ring_doorbell")
version = _version("ring_doorbell")

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "myst_parser",
]

myst_enable_extensions = [
    "colon_fence",
]

templates_path = ["_templates"]
exclude_patterns: list[str] = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
master_doc = "index"



"""The tests for the Ring platform."""

import datetime
import json

import pytest
from freezegun.api import FrozenDateTimeFactory
from ring_doorbell import Ring
from ring_doorbell.exceptions import RingError
from ring_doorbell.listen import RingEventListener

from tests.conftest import load_alert_v1, load_alert_v2, load_fixture


async def test_listen(auth, mocker):
    import firebase_messaging

    ring = Ring(auth)
    listener = RingEventListener(ring)

    await listener.start()
    assert firebase_messaging.FcmPushClient.checkin_or_register.call_count == 1
    assert firebase_messaging.FcmPushClient.start.call_count == 1
    assert listener.subscribed is True
    assert listener.started is True

    with pytest.raises(RingError, match="ID 10 is not a valid callback id"):
        listener.remove_notification_callback(10)

    with pytest.raises(
        RingError,
        match="Cannot remove the default callback for ring-doorbell with value 1",
    ):
        listener.remove_notification_callback(1)

    cbid = listener.add_notification_callback(lambda: 2)
    del listener._callbacks[1]
    listener.remove_notification_callback(cbid)


async def test_active_dings(auth, mocker):
    import firebase_messaging

    ring = Ring(auth)
    listener = RingEventListener(ring)
    await listener.start()
    assert firebase_messaging.FcmPushClient.checkin_or_register.call_count == 1
    assert firebase_messaging.FcmPushClient.start.call_count == 1
    assert listener.subscribed is True
    assert listener.started is True
    num_active = len(ring.active_alerts())
    assert num_active == 0
    alertstoadd = 2
    for i in range(alertstoadd):
        msg = load_alert_v1("doorbot_ding", 123456781, ding_id_inc=i)
        listener._on_notification(msg, "1234567" + str(i))
        msg = load_alert_v2("camera_motion", 123456782, ding_id_inc=i)
        listener._on_notification(msg, "1234567" + str(i))
        msg = load_alert_v1("intercom_unlock", 185036587, ding_id_inc=i)
        listener._on_notification(msg, "1234567" + str(i))

    dings = ring.active_alerts()
    assert len(dings) == num_active + alertstoadd * 3
    # Test with the same id which should overwrite
    # previous and keep the overall count the same
    for i in range(alertstoadd):
        msg = load_alert_v1("doorbot_ding", 123456781, ding_id_inc=i)
        listener._on_notification(msg, "1234567" + str(i))
        msg = load_alert_v2("camera_motion", 123456782, ding_id_inc=i)
        listener._on_notification(msg, "1234567" + str(i))

    dings = ring.active_alerts()
    assert len(dings) == num_active + alertstoadd * 3
    await listener.stop()


async def test_ding_expirey(auth, mocker, freezer: FrozenDateTimeFactory):
    ring = Ring(auth)
    listener = RingEventListener(ring)
    await listener.start()

    assert listener.subscribed is True
    assert listener.started is True

    assert len(ring.push_dings_data) == 0
    assert len(ring.active_alerts()) == 0

    alertstoadd = 2
    for i in range(alertstoadd):
        msg = load_alert_v1("doorbot_ding", 123456781, ding_id_inc=i)
        listener._on_notification(msg, "1234567" + str(i))
        msg = load_alert_v2("camera_motion", 123456782, ding_id_inc=i)
        listener._on_notification(msg, "1234567" + str(i))
        msg = load_alert_v1("intercom_unlock", 185036587, ding_id_inc=i)
        listener._on_notification(msg, "1234567" + str(i))

    assert len(ring.push_dings_data) == 6
    assert len(ring.active_alerts()) == 6

    freezer.tick(datetime.timedelta(minutes=5))

    msg = load_alert_v1("doorbot_ding", 123456781, ding_id_inc=alertstoadd + 1)
    listener._on_notification(msg, "123456781" + str(alertstoadd + 1))

    assert len(ring.push_dings_data) == 1
    assert len(ring.active_alerts()) == 1


@pytest.mark.nolistenmock
async def test_listen_subscribe_fail(auth, mocker, caplog, putpatch_status_fixture):
    checkinmock = mocker.patch(
        "firebase_messaging.FcmPushClient.checkin_or_register", return_value="foobar"
    )
    connectmock = mocker.patch("firebase_messaging.FcmPushClient.start")
    mocker.patch("firebase_messaging.FcmPushClient.is_started", return_value=True)

    putpatch_status_fixture.overrides["https://api.ring.com/clients_api/device"] = 401

    ring = Ring(auth)
    listener = RingEventListener(ring)
    await listener.start()
    # Check in gets and error so register is called
    assert checkinmock.call_count == 1
    assert listener.subscribed is False
    assert listener.started is False
    assert connectmock.call_count == 0

    exp = (
        "Unable to checkin to listen service, "
        "response was 401 , event listener not started"
    )
    assert (
        len(
            [
                record
                for record in caplog.records
                if record.levelname == "ERROR" and record.message == exp
            ]
        )
        == 1
    )


@pytest.mark.nolistenmock
async def test_listen_gcm_fail(auth, mocker):
    # Check in gets and error so register is called, the subscribe gets an error
    credentials = json.loads(load_fixture("ring_listen_credentials.json"))
    checkinmock = mocker.patch(
        "firebase_messaging.fcmregister.FcmRegister.gcm_check_in", return_value=None
    )
    registermock = mocker.patch(
        "firebase_messaging.fcmregister.FcmRegister.register", return_value=credentials
    )
    connectmock = mocker.patch("firebase_messaging.FcmPushClient.start")
    mocker.patch("firebase_messaging.FcmPushClient.is_started", return_value=True)

    ring = Ring(auth)
    listener = RingEventListener(ring, credentials)
    await listener.start()
    # Check in gets and error so register is called
    assert checkinmock.call_count == 1
    assert registermock.call_count == 1
    assert listener.subscribed is True
    assert listener.started is True
    assert connectmock.call_count == 1


@pytest.mark.nolistenmock
async def test_listen_fcm_fail(auth, mocker, caplog):
    checkinmock = mocker.patch(
        "firebase_messaging.FcmPushClient.checkin_or_register", return_value=None
    )
    connectmock = mocker.patch("firebase_messaging.FcmPushClient.start")
    mocker.patch("firebase_messaging.FcmPushClient.is_started", return_value=True)

    ring = Ring(auth)
    listener = RingEventListener(ring)
    await listener.start()
    # Check in gets and error so register is called
    assert checkinmock.call_count == 1

    assert listener.subscribed is False
    assert listener.started is False
    assert connectmock.call_count == 0
    exp = "Ring listener unable to check in to fcm, event listener not started"
    assert (
        len(
            [
                record
                for record in caplog.records
                if record.levelname == "ERROR" and record.message == exp
            ]
        )
        == 1
    )



"""Module for cli tests."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any
from unittest.mock import DEFAULT, MagicMock, call, patch

import aiofiles
import pytest
from asyncclick.testing import CliRunner
from ring_doorbell import AuthenticationError, Requires2FAError, Ring
from ring_doorbell.cli import (
    _event_handler,
    cli,
    devices_command,
    in_home_chime,
    list_command,
    listen,
    motion_detection,
    open_door,
    show,
    videos,
)
from ring_doorbell.const import GCM_TOKEN_FILE

from tests.conftest import (
    load_alert_v1,
    load_fixture,
    load_fixture_as_dict,
)


async def test_cli_default(ring):
    runner = CliRunner()
    with runner.isolated_filesystem():
        res = await runner.invoke(
            cli, ["--username", "foo", "--password", "foo"], obj=ring
        )

        expected = (
            "Name:       Downstairs\nFamily:     chimes\nID:"
            "         999999\nTimezone:   America/New_York\nWifi Name:"
            "  ring_mock_wifi\nWifi RSSI:  -39\n\n"
        )

        assert res.exit_code == 0
        assert expected in res.output


async def test_show(ring):
    runner = CliRunner()
    with runner.isolated_filesystem():
        res = await runner.invoke(show, obj=ring)

        expected = (
            "Name:       Downstairs\nFamily:     chimes\nID:"
            "         999999\nTimezone:   America/New_York\nWifi Name:"
            "  ring_mock_wifi\nWifi RSSI:  -39\n\n"
        )

        assert res.exit_code == 0
        assert expected in res.output


async def test_devices(ring):
    runner = CliRunner()
    with runner.isolated_filesystem():
        res = await runner.invoke(devices_command, obj=ring)

        expected = (
            "(Pretty format coming soon, if you want json "
            "consistently from this command provide the --json flag)\n"
        )
        for device_type in ring.devices_data:
            for device_api_id in ring.devices_data[device_type]:
                expected += (
                    json.dumps(ring.devices_data[device_type][device_api_id], indent=2)
                    + "\n"
                )

        assert res.exit_code == 0
        assert expected == res.output


async def test_list(ring):
    runner = CliRunner()
    with runner.isolated_filesystem():
        res = await runner.invoke(list_command, obj=ring)

        expected = (
            "Front Door (lpd_v1)\nBack Door (lpd_v1)\nDownstairs (chime)\n"
            "Front (hp_cam_v1)\nIngress (intercom_handset_audio)\n"
        )

        assert res.exit_code == 0
        assert expected in res.output


aiofiles.threadpool.wrap.register(MagicMock)(
    lambda *args, **kwargs: aiofiles.threadpool.AsyncBufferedIOBase(*args, **kwargs)
)


async def test_videos(ring, mocker):
    runner = CliRunner()
    mock_file_stream = MagicMock(read=lambda *args, **kwargs: b"")  # noqa: ARG005
    with runner.isolated_filesystem():
        with patch(
            "aiofiles.threadpool.sync_open", return_value=mock_file_stream
        ) as mock_open:
            res = await runner.invoke(videos, ["--count", "--download-all"], obj=ring)
        assert mock_open.call_count == 3
        mock_file_stream.write.assert_has_calls([call(b"123456")] * 3)
        assert "Downloading 3 videos" in res.output


@pytest.mark.parametrize(
    ("affect_method", "exception", "file_exists"),
    [
        (None, None, False),
        ("ring_doorbell.auth.Auth.async_fetch_token", Requires2FAError, False),
        ("ring_doorbell.ring.Ring.async_update_data", AuthenticationError, True),
    ],
    ids=("No 2FA", "Require 2FA", "Invalid Grant"),
)
async def test_auth(mocker, affect_method, exception, file_exists):
    call_count = 0

    def _raise_once(self, *_: dict[str, Any], **__: dict[str, Any]) -> dict[str, Any]:
        nonlocal call_count, exception
        if call_count == 0:
            call_count += 1
            msg = "Simulated exception"
            raise exception(msg)
        call_count += 1

        if hasattr(self, "_update_data"):
            return self._update_data()

        return DEFAULT

    def _add_token(
        uri,
        http_method,
        body,
        headers,
        token_placement=None,
        **kwargs,  # noqa: ANN003
    ) -> tuple[str, dict, bytes]:
        return uri, headers, body

    mocker.patch(
        "oauthlib.oauth2.LegacyApplicationClient.add_token", side_effect=_add_token
    )
    mocker.patch.object(Path, "is_file", return_value=file_exists)
    mocker.patch.object(Path, "read_text", return_value=load_fixture("ring_oauth.json"))
    mocker.patch("builtins.input", return_value="Foo")
    mocker.patch("getpass.getpass", return_value="Foo")
    if affect_method is not None:
        mocker.patch(affect_method, side_effect=_raise_once, autospec=True)

    runner = CliRunner()
    with runner.isolated_filesystem():
        res = await runner.invoke(cli)

        assert res.exit_code == 0


async def test_motion_detection(ring, aioresponses_mock, devices_fixture):
    runner = CliRunner()
    with runner.isolated_filesystem():
        res = await runner.invoke(
            motion_detection,
            ["--device-name", "Front"],
            obj=ring,
        )
        expected = "Front (hp_cam_v1) has motion detection off"
        assert res.exit_code == 0
        assert expected in res.output

        res = await runner.invoke(
            motion_detection,
            ["--device-name", "Front", "--off"],
            obj=ring,
        )
        expected = "Front (hp_cam_v1) already has motion detection off"
        assert res.exit_code == 0
        assert expected in res.output

        # Changes the return to indicate that the siren is now on.
        devices_fixture.updated = True
        aioresponses_mock.get(
            "https://api.ring.com/clients_api/ring_devices",
            payload=load_fixture_as_dict("ring_devices_updated.json"),
        )

        res = await runner.invoke(
            motion_detection,
            ["--device-name", "Front", "--on"],
            obj=ring,
        )
        expected = "Front (hp_cam_v1) motion detection set to on"
        assert res.exit_code == 0
        assert expected in res.output


@pytest.mark.nolistenmock
async def test_listen_store_credentials(mocker, auth):
    runner = CliRunner()
    import firebase_messaging

    credentials = json.loads(load_fixture("ring_listen_credentials.json"))

    with runner.isolated_filesystem():
        mocker.patch(
            "firebase_messaging.fcmregister.FcmRegister.gcm_check_in",
            return_value="foobar",
        )
        mocker.patch(
            "firebase_messaging.fcmregister.FcmRegister.register",
            return_value=credentials,
        )
        mocker.patch("firebase_messaging.FcmPushClient.start")
        mocker.patch("firebase_messaging.FcmPushClient.is_started", return_value=True)

        ring = Ring(auth)
        assert not Path(GCM_TOKEN_FILE).is_file()

        await runner.invoke(
            listen, ["--store-credentials"], obj=ring, catch_exceptions=True
        )
        assert Path(GCM_TOKEN_FILE).is_file()
        assert firebase_messaging.fcmregister.FcmRegister.gcm_check_in.call_count == 0
        assert firebase_messaging.fcmregister.FcmRegister.register.call_count == 1
        assert firebase_messaging.FcmPushClient.start.call_count == 1

        ring = Ring(auth)
        await runner.invoke(listen, ["--store-credentials"], obj=ring)
        assert firebase_messaging.fcmregister.FcmRegister.gcm_check_in.call_count == 1
        assert firebase_messaging.fcmregister.FcmRegister.register.call_count == 1
        assert firebase_messaging.FcmPushClient.start.call_count == 2


async def test_listen_event_handler(mocker, auth):
    from ring_doorbell.listen import RingEventListener

    ring = Ring(auth)
    listener = RingEventListener(ring)
    await listener.start()
    listener.add_notification_callback(_event_handler(ring).on_event)

    msg = load_alert_v1(
        "camera_motion", 12345678, created_at="2023-10-24T09:42:18.789709Z"
    )
    echomock = mocker.patch("ring_doorbell.cli.echo")
    mocker.patch(
        "ring_doorbell.cli.get_now_str", return_value="2023-10-24 09:42:18.789709"
    )
    listener._on_notification(msg, "1234567")
    exp = (
        "2023-10-24 09:42:18.789709: RingEvent(id=12345678901234, "
        "doorbot_id=12345678, device_name='Front Floodcam'"
        ", device_kind='floodlight_v2', now=1698140538.789709,"
        " expires_in=180, kind='motion', state='human', is_update=False) : "
        "Currently active count = 1"
    )
    echomock.assert_called_with(exp)


async def test_in_home_chime(ring, aioresponses_mock, devices_fixture):
    runner = CliRunner()
    with runner.isolated_filesystem():
        # Gets in-home chime details for a doorbell
        res = await runner.invoke(
            in_home_chime,
            ["--device-name", "Front Door"],
            obj=ring,
        )
        expected_show = (
            "Name:       Front Door\n"
            "ID:         987652\n"
            "Type:       Mechanical\n"
            "Enabled:    True\n"
            "Duration:   None\n"
        )
        expected = expected_show
        assert res.exit_code == 0
        assert expected in res.output

        # Turns off the in-home chime
        res = await runner.invoke(
            in_home_chime,
            ["--device-name", "Front Door", "enabled", "False"],
            obj=ring,
        )
        expected = "Front Door's in-home chime has been disabled"
        assert res.exit_code == 0
        assert expected in res.output

        # Turns on the in-home chime and sets type to Mechanical
        res = await runner.invoke(
            in_home_chime,
            [
                "--device-name",
                "Front Door",
                "type",
                "mechanical",
            ],
            obj=ring,
        )
        expected = "Front Door's in-home chime type has been set to Mechanical\n"
        assert res.exit_code == 0
        assert expected in res.output

        # Sets type to Digital and changes the duration
        res = await runner.invoke(
            in_home_chime,
            ["--device-name", "Front Door", "duration", "5"],
            obj=ring,
        )
        expected = "Front Door's in-home chime duration has been set to 5 seconds\n"
        assert res.exit_code == 0
        assert expected in res.output

        # Runs in-home-chime against a device that doesn't have an in-home chime
        res = await runner.invoke(
            in_home_chime,
            ["--device-name", "Front"],
            obj=ring,
        )
        expected = "Front is not a doorbell"
        assert res.exit_code == 1
        assert expected in res.output

        # Runs in-home-chime with no parameters and error as more than one doorbot.
        res = await runner.invoke(
            in_home_chime,
            [],
            obj=ring,
        )
        expected = "There are 2 doorbells, you need to pass the --device-name option"
        assert res.exit_code == 1
        assert expected in res.output


async def test_open_door(ring, aioresponses_mock, devices_fixture):
    runner = CliRunner()

    res = await runner.invoke(
        open_door,
        ["--device-name", "Ingress"],
        obj=ring,
    )
    assert res.exit_code == 0
    assert res.output == "Ingress opened\n"


async def test_get_device(ring, aioresponses_mock, devices_fixture):
    runner = CliRunner()

    # Get device by name
    res = await runner.invoke(
        open_door,
        ["--device-name", "Ingress"],
        obj=ring,
    )
    assert res.exit_code == 0
    assert res.output == "Ingress opened\n"

    # Get device by single type
    res = await runner.invoke(
        open_door,
        [],
        obj=ring,
    )
    assert res.exit_code == 0
    assert res.output == "Ingress opened\n"

    # Get wrong device type
    res = await runner.invoke(
        open_door,
        ["--device-name", "Front"],
        obj=ring,
    )
    assert res.exit_code == 1
    assert "Front is not a intercom" in res.output

    # Wrong name
    res = await runner.invoke(
        open_door,
        ["--device-name", "Frontx"],
        obj=ring,
    )
    assert res.exit_code == 1
    assert "Cannot find intercom with name Frontx" in res.output



"""Module for RingEventListenerConfig."""

from __future__ import annotations

from firebase_messaging import FcmPushClientConfig


class RingEventListenerConfig(FcmPushClientConfig):
    """Configuration class for event listener."""

    @staticmethod
    def default_config() -> RingEventListenerConfig:
        """Get an instance of the default config."""
        config = RingEventListenerConfig()
        config.server_heartbeat_interval = 60
        config.client_heartbeat_interval = 120
        config.monitor_interval = 15
        return config



# vim:sw=4:ts=4:et:
"""Python Ring Doorbell wrapper."""

from __future__ import annotations

import asyncio
import logging
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

import aiofiles

from ring_doorbell.const import (
    DEFAULT_VIDEO_DOWNLOAD_TIMEOUT,
    DINGS_ENDPOINT,
    DOORBELL_2_KINDS,
    DOORBELL_3_KINDS,
    DOORBELL_3_PLUS_KINDS,
    DOORBELL_4_KINDS,
    DOORBELL_BATTERY_KINDS,
    DOORBELL_ELITE_KINDS,
    DOORBELL_EXISTING_DURATION_MAX,
    DOORBELL_EXISTING_DURATION_MIN,
    DOORBELL_EXISTING_TYPE,
    DOORBELL_GEN2_KINDS,
    DOORBELL_KINDS,
    DOORBELL_PRO_2_KINDS,
    DOORBELL_PRO_KINDS,
    DOORBELL_VOL_MAX,
    DOORBELL_VOL_MIN,
    DOORBELL_WIRED_KINDS,
    DOORBELLS_ENDPOINT,
    FILE_EXISTS,
    HEALTH_DOORBELL_ENDPOINT,
    ICE_SERVERS,
    LIVE_STREAMING_ENDPOINT,
    MSG_ALLOWED_VALUES,
    MSG_BOOLEAN_REQUIRED,
    MSG_EXISTING_TYPE,
    MSG_EXPECTED_ATTRIBUTE_NOT_FOUND,
    MSG_VOL_OUTBOUND,
    PEEPHOLE_CAM_KINDS,
    SETTINGS_ENDPOINT,
    SNAPSHOT_ENDPOINT,
    SNAPSHOT_TIMESTAMP_ENDPOINT,
    URL_RECORDING,
    URL_RECORDING_SHARE_PLAY,
    RingCapability,
)
from ring_doorbell.exceptions import RingError
from ring_doorbell.generic import RingGeneric
from ring_doorbell.webrtcstream import RingWebRtcMessageCallback, RingWebRtcStream

_LOGGER = logging.getLogger(__name__)


class RingDoorBell(RingGeneric):
    """Implementation for Ring Doorbell."""

    if TYPE_CHECKING:
        from ring_doorbell.ring import Ring

    def __init__(self, ring: Ring, device_api_id: int, *, shared: bool = False) -> None:
        """Initialise the doorbell."""
        super().__init__(ring, device_api_id)
        self.shared = shared
        self._webrtc_streams: dict[str, RingWebRtcStream] = {}

    @property
    def family(self) -> str:
        """Return Ring device family type."""
        return "authorized_doorbots" if self.shared else "doorbots"

    async def async_update_health_data(self) -> None:
        """Update health attrs."""
        resp = await self._ring.async_query(
            HEALTH_DOORBELL_ENDPOINT.format(self.device_api_id)
        )
        self._health_attrs = resp.json().get("device_health", {})

    @property
    def model(self) -> str:  # noqa: C901, PLR0911
        """Return Ring device model name."""
        if self.kind in DOORBELL_KINDS:
            return "Doorbell"
        if self.kind in DOORBELL_2_KINDS:
            return "Doorbell 2"
        if self.kind in DOORBELL_3_KINDS:
            return "Doorbell 3"
        if self.kind in DOORBELL_3_PLUS_KINDS:
            return "Doorbell 3 Plus"
        if self.kind in DOORBELL_4_KINDS:
            return "Doorbell 4"
        if self.kind in DOORBELL_PRO_KINDS:
            return "Doorbell Pro"
        if self.kind in DOORBELL_PRO_2_KINDS:
            return "Doorbell Pro 2"
        if self.kind in DOORBELL_ELITE_KINDS:
            return "Doorbell Elite"
        if self.kind in DOORBELL_WIRED_KINDS:
            return "Doorbell Wired"
        if self.kind in DOORBELL_BATTERY_KINDS:
            return "Battery Doorbell"
        if self.kind in DOORBELL_GEN2_KINDS:
            return "Doorbell (2nd Gen)"
        if self.kind in PEEPHOLE_CAM_KINDS:
            return "Peephole Cam"
        return "Unknown Doorbell"

    def has_capability(self, capability: RingCapability | str) -> bool:  # noqa: PLR0911
        """Return if device has specific capability."""
        capability = (
            capability
            if isinstance(capability, RingCapability)
            else RingCapability.from_name(capability)
        )
        if capability == RingCapability.BATTERY:
            return self.kind in (
                DOORBELL_KINDS
                + DOORBELL_2_KINDS
                + DOORBELL_3_KINDS
                + DOORBELL_3_PLUS_KINDS
                + DOORBELL_4_KINDS
                + DOORBELL_GEN2_KINDS
                + DOORBELL_BATTERY_KINDS
                + PEEPHOLE_CAM_KINDS
            )
        if capability == RingCapability.KNOCK:
            return self.kind in PEEPHOLE_CAM_KINDS
        if capability == RingCapability.PRE_ROLL:
            return self.kind in DOORBELL_3_PLUS_KINDS
        if capability == RingCapability.VOLUME:
            return True
        if capability == RingCapability.HISTORY:
            return True
        if capability in [
            RingCapability.MOTION_DETECTION,
            RingCapability.VIDEO,
            RingCapability.DING,
        ]:
            return self.kind in (
                DOORBELL_KINDS
                + DOORBELL_2_KINDS
                + DOORBELL_3_KINDS
                + DOORBELL_3_PLUS_KINDS
                + DOORBELL_4_KINDS
                + DOORBELL_PRO_KINDS
                + DOORBELL_PRO_2_KINDS
                + DOORBELL_WIRED_KINDS
                + DOORBELL_BATTERY_KINDS
                + DOORBELL_GEN2_KINDS
                + DOORBELL_ELITE_KINDS
                + PEEPHOLE_CAM_KINDS
            )
        return False

    @property
    def battery_life(self) -> int | None:
        """Return battery life."""
        if (
            bl1 := self._attrs.get("battery_life")
        ) is None and "battery_life_2" not in self._attrs:
            return None

        value = 0
        if bl1:
            value += int(bl1)

        if bl2 := self._attrs.get("battery_life_2"):  # Camera has two battery bays
            value += int(bl2)

        return min(value, 100)

    def _get_chime_setting(self, setting: str) -> Any | None:
        if (settings := self._attrs.get("settings")) and (
            chime_settings := settings.get("chime_settings")
        ):
            return chime_settings.get(setting)
        return None

    @property
    def existing_doorbell_type(self) -> str | None:
        """
        Return existing doorbell type.

        0: Mechanical
        1: Digital
        2: Not Present
        """
        try:
            if (dtype := self._get_chime_setting("type")) is not None:
                return DOORBELL_EXISTING_TYPE[dtype]
        except AttributeError:
            return None
        else:
            return None

    async def async_set_existing_doorbell_type(self, value: int) -> None:
        """
        Return existing doorbell type.

        0: Mechanical
        1: Digital
        2: Not Present
        """
        if value not in DOORBELL_EXISTING_TYPE:
            msg = f"value must be in {MSG_EXISTING_TYPE}"
            raise RingError(msg)
        params = {
            "doorbot[description]": self.name,
            "doorbot[settings][chime_settings][type]": value,
        }
        if self.existing_doorbell_type:
            url = DOORBELLS_ENDPOINT.format(self.device_api_id)
            await self._ring.async_query(url, extra_params=params, method="PUT")

    @property
    def existing_doorbell_type_enabled(self) -> bool | None:
        """Return if existing doorbell type is enabled."""
        if self.existing_doorbell_type:
            if self.existing_doorbell_type == DOORBELL_EXISTING_TYPE[2]:
                return None
            return self._get_chime_setting("enable")
        return False

    async def async_set_existing_doorbell_type_enabled(self, value: bool) -> None:  # noqa: FBT001
        """Enable/disable the existing doorbell if Digital/Mechanical."""
        if self.existing_doorbell_type:
            if not isinstance(value, bool):
                raise RingError(MSG_BOOLEAN_REQUIRED)

            if self.existing_doorbell_type == DOORBELL_EXISTING_TYPE[2]:
                msg = "In-Home chime is not present."
                raise RingError(msg)

            int_value = int(value)

            params = {
                "doorbot[description]": self.name,
                "doorbot[settings][chime_settings][enable]": int_value,
            }
            url = DOORBELLS_ENDPOINT.format(self.device_api_id)
            await self._ring.async_query(url, extra_params=params, method="PUT")

    @property
    def existing_doorbell_type_duration(self) -> int | None:
        """Return duration for Digital chime."""
        if (
            self.existing_doorbell_type
            and self.existing_doorbell_type == DOORBELL_EXISTING_TYPE[1]
        ):
            return self._get_chime_setting("duration")
        return None

    async def async_set_existing_doorbell_type_duration(self, value: int) -> None:
        """Set duration for Digital chime."""
        if self.existing_doorbell_type:
            if not (
                (isinstance(value, int))
                and (
                    DOORBELL_EXISTING_DURATION_MIN
                    <= value
                    <= DOORBELL_EXISTING_DURATION_MAX
                )
            ):
                raise RingError(
                    MSG_VOL_OUTBOUND.format(
                        DOORBELL_EXISTING_DURATION_MIN, DOORBELL_EXISTING_DURATION_MAX
                    )
                )

            if self.existing_doorbell_type == DOORBELL_EXISTING_TYPE[1]:
                params = {
                    "doorbot[description]": self.name,
                    "doorbot[settings][chime_settings][duration]": value,
                }
                url = DOORBELLS_ENDPOINT.format(self.device_api_id)
                await self._ring.async_query(url, extra_params=params, method="PUT")

    async def async_get_last_recording_id(self) -> int | None:
        """Return the last recording ID."""
        try:
            res = await self.async_history(limit=1)
            return res[0].get("id") if res else None
        except (IndexError, TypeError):
            return None

    async def async_get_live_streaming_json(self) -> dict[str, Any] | None:
        """Return JSON for live streaming."""
        url = LIVE_STREAMING_ENDPOINT.format(self.device_api_id)
        req = await self._ring.async_query(
            url, method="POST", base_uri="https://app.ring.com"
        )
        if req and req.status_code == 200:
            url = DINGS_ENDPOINT
            try:
                resp = await self._ring.async_query(url)
                return resp.json()[0]
            except (IndexError, TypeError):
                pass
        return None

    async def async_recording_download(
        self,
        recording_id: int,
        filename: str | None = None,
        *,
        override: bool = False,
        timeout: int = DEFAULT_VIDEO_DOWNLOAD_TIMEOUT,
    ) -> bytes | None:
        """Save a recording in MP4 format to a file or return raw."""
        if not self.has_subscription:
            msg = "Your Ring account does not have an active subscription."
            _LOGGER.warning(msg)
            return None

        url = URL_RECORDING.format(recording_id)
        try:
            # Video download needs a longer timeout to get the large video file
            req = await self._ring.async_query(url, timeout=timeout)
            if req.status_code == 200:
                if filename:
                    if Path(filename).is_file() and not override:
                        raise RingError(FILE_EXISTS.format(filename))

                    async with aiofiles.open(filename, "wb") as recording:
                        await recording.write(req.content)
                        return None
                else:
                    return req.content
            else:
                msg = (
                    f"Could not get recording at url {url}, "
                    f"status code is {req.status_code}"
                )
                raise RingError(msg)
        except OSError as error:
            msg = f"Error downloading recording {recording_id}: {error}"
            _LOGGER.exception(msg)
            raise RingError(msg) from error

    async def async_recording_url(self, recording_id: int) -> str | None:
        """Return HTTPS recording URL."""
        if not self.has_subscription:
            msg = "Your Ring account does not have an active subscription."
            _LOGGER.warning(msg)
            return None

        url = URL_RECORDING_SHARE_PLAY.format(recording_id)
        req = await self._ring.async_query(url)
        data = req.json()
        if req and req.status_code == 200 and data is not None:
            return data["url"]
        return None

    @property
    def subscribed(self) -> bool:
        """Return if is online."""
        result = self._attrs.get("subscribed")
        return result is not None

    @property
    def subscribed_motion(self) -> bool:
        """Return if is subscribed_motion."""
        result = self._attrs.get("subscribed_motions")
        return result is not None

    @property
    def has_subscription(self) -> bool:
        """Return boolean if the account has subscription."""
        if features := self._attrs.get("features"):
            return features.get("show_recordings", False)
        return False

    @property
    def volume(self) -> int:
        """Return the volume."""
        return self._attrs["settings"].get("doorbell_volume", 0)

    async def async_set_volume(self, value: int) -> None:
        """Set the volume."""
        if not (
            (isinstance(value, int)) and (DOORBELL_VOL_MIN <= value <= DOORBELL_VOL_MAX)
        ):
            raise RingError(MSG_VOL_OUTBOUND.format(DOORBELL_VOL_MIN, DOORBELL_VOL_MAX))

        params = {
            "doorbot[description]": self.name,
            "doorbot[settings][doorbell_volume]": str(value),
        }
        url = DOORBELLS_ENDPOINT.format(self.device_api_id)
        await self._ring.async_query(url, extra_params=params, method="PUT")

    @property
    def connection_status(self) -> str | None:
        """Return connection status."""
        if alerts := self._attrs.get("alerts"):
            return alerts.get("connection")
        return None

    async def async_get_snapshot(
        self, retries: int = 3, delay: int = 1, filename: str | None = None
    ) -> bytes | None:
        """Take a snapshot and download it."""
        url = SNAPSHOT_TIMESTAMP_ENDPOINT
        payload = {"doorbot_ids": [self._attrs.get("id")]}
        await self._ring.async_query(url, method="POST", json=payload)
        request_time = time.time()
        for _ in range(retries):
            await asyncio.sleep(delay)
            resp = await self._ring.async_query(url, method="POST", json=payload)
            response = resp.json()
            if response["timestamps"][0]["timestamp"] / 1000 > request_time:
                resp = await self._ring.async_query(
                    SNAPSHOT_ENDPOINT.format(self._attrs.get("id"))
                )
                snapshot = resp.content
                if filename:
                    async with aiofiles.open(filename, "wb") as jpg:
                        await jpg.write(snapshot)
                    return None
                return snapshot
        return None

    def _motion_detection_state(self) -> bool | None:
        if settings := self._attrs.get("settings"):
            return settings.get("motion_detection_enabled")
        return None

    @property
    def motion_detection(self) -> bool:
        """Return motion detection enabled state."""
        return state if (state := self._motion_detection_state()) else False

    async def async_set_motion_detection(self, state: bool) -> None:  # noqa: FBT001
        """Set the motion detection enabled state."""
        values = [True, False]
        if state not in values:
            raise RingError(MSG_ALLOWED_VALUES.format("True, False"))

        if self._motion_detection_state() is None:
            _LOGGER.warning(
                "%s",
                MSG_EXPECTED_ATTRIBUTE_NOT_FOUND.format(
                    "settings[motion_detection_enabled]"
                ),
            )
            return

        url = SETTINGS_ENDPOINT.format(self.device_api_id)
        payload = {"motion_settings": {"motion_detection_enabled": state}}

        await self._ring.async_query(url, method="PATCH", json=payload)

    async def generate_webrtc_stream(
        self, sdp_offer: str, *, keep_alive_timeout: int | None = 30
    ) -> str:
        """Generate the rtc stream."""
        if session_id := RingWebRtcStream.get_sdp_session_id(sdp_offer):

            async def _close_callback() -> None:
                await self.close_webrtc_stream(session_id)

            stream = RingWebRtcStream(
                self._ring,
                self.device_api_id,
                keep_alive_timeout=keep_alive_timeout,
                on_close_callback=_close_callback,
            )
            sdp_answer = await stream.generate(sdp_offer)
            # generate will raise if no sdp answer as no callback passed
            assert sdp_answer  # noqa: S101
            self._webrtc_streams[session_id] = stream
            return sdp_answer
        msg = "Unable to generate the stream, could not extract session id from offer."
        raise RingError(msg)

    async def generate_async_webrtc_stream(
        self,
        sdp_offer: str,
        session_id: str,
        on_message_callback: RingWebRtcMessageCallback,
        *,
        keep_alive_timeout: int | None = 60 * 5,
    ) -> None:
        """Generate the rtc stream. Will callback with answers and ICE candidates."""

        async def _close_callback() -> None:
            await self.close_webrtc_stream(session_id)

        stream = RingWebRtcStream(
            self._ring,
            self.device_api_id,
            on_message_callback=on_message_callback,
            keep_alive_timeout=keep_alive_timeout,
            on_close_callback=_close_callback,
        )
        self._webrtc_streams[session_id] = stream
        await stream.generate(sdp_offer)

    async def on_webrtc_candidate(
        self, session_id: str, candidate: str, multi_line_index: int
    ) -> None:
        """Send an ICE candidate."""
        if stream := self._webrtc_streams.get(session_id):
            await stream.on_ice_candidate(candidate, multi_line_index)
        else:
            msg = "Ice candidate received before stream has been created."
            raise RingError(msg)

    async def close_webrtc_stream(self, session_id: str) -> None:
        """Close the rtc stream."""
        stream = self._webrtc_streams.pop(session_id, None)
        if stream:
            await stream.close()

    def sync_close_webrtc_stream(self, session_id: str) -> None:
        """Close the rtc stream."""
        stream = self._webrtc_streams.pop(session_id, None)
        if stream:
            stream.sync_close()

    async def keep_alive_webrtc_stream(self, sdp_session_id: str) -> None:
        """Keep alive the rtc stream."""
        stream = self._webrtc_streams.get(sdp_session_id, None)
        if stream:
            await stream.keep_alive()

    def get_ice_servers(self) -> list[str]:
        """Return the ICE servers."""
        return ICE_SERVERS

    DEPRECATED_API_QUERIES: ClassVar = {
        *RingGeneric.DEPRECATED_API_QUERIES,
        "update_health_data",
        "recording_download",
        "recording_url",
        "get_snapshot",
    }
    DEPRECATED_API_PROPERTY_GETTERS: ClassVar = {
        *RingGeneric.DEPRECATED_API_PROPERTY_GETTERS,
        "last_recording_id",
        "live_streaming_json",
    }
    DEPRECATED_API_PROPERTY_SETTERS: ClassVar = {
        *RingGeneric.DEPRECATED_API_PROPERTY_SETTERS,
        "existing_doorbell_type",
        "existing_doorbell_type_enabled",
        "existing_doorbell_type_duration",
        "volume",
        "motion_detection",
    }
